# %%
import timeit

import dask
import eofs
import numpy as np
import xarray as xr
from tqdm import tqdm

import xeofs as xe


# %%
def fit_eofs(X, n_modes=2):
    _ = eofs.xarray.Eof(X)


def fit_xeofs(X, n_modes=2):
    model = xe.single.EOF(n_modes=n_modes, random_state=5)
    model.fit(X, dim="time")


def create_test_data(ns, nf1, nf2, delayed=True, seed=None):
    size = ns * nf1 * nf2
    alpha = size / 1e7
    n_chunks = int(alpha) + 1 if alpha > 1 else 1
    chunk_size = int(np.ceil(ns / n_chunks))

    rng = dask.array.random.default_rng(seed)
    X = rng.standard_normal(
        (ns, nf1, nf2), chunks=(chunk_size, -1, -1), dtype=np.float32
    )
    X = xr.DataArray(
        X,
        dims=("time", "lon", "lat"),
        coords={
            "time": np.arange(ns),
            "lon": np.arange(nf1),
            "lat": np.arange(nf2),
        },
        name="test_data",
    )
    if delayed:
        return X
    else:
        return X.load()


# %%

MAX_MEMORY = {"eofs": 1000, "xeofs": 3000}
n_repeats = 3

n_samples = np.array([10, 20, 60, 100, 200, 600, 1_000, 2_000, 6_000, 10_000])[::-1]
n_features_x = np.array(
    [1, 2, 6, 10, 20, 60, 100, 200, 600, 1_000, 2_000, 6_000, 10_000]
)[::-1]

timings = {
    "data": np.empty((len(n_samples), len(n_features_x))) * np.nan,
    "eofs": np.empty((len(n_samples), len(n_features_x), n_repeats)) * np.nan,
    "xeofs": np.empty((len(n_samples), len(n_features_x), n_repeats)) * np.nan,
    "eofs_dask": np.empty((len(n_samples), len(n_features_x), n_repeats)) * np.nan,
    "xeofs_dask": np.empty((len(n_samples), len(n_features_x), n_repeats)) * np.nan,
}

for i, ns in enumerate(n_samples):
    print(f"n_samples = {ns}")
    for j, nf in tqdm(enumerate(n_features_x), total=len(n_features_x)):
        # Create test data
        X = create_test_data(ns, nf, 10, delayed=True, seed=8)
        nbytes = X.nbytes / 1e6  # in MiB

        # Save data size
        timings["data"][i][j] = X.nbytes

        # Set number of loops
        if 10 < nbytes:
            n_loops = 1
        elif 1 < nbytes and nbytes <= 10:
            n_loops = 10
        elif nbytes <= 1:
            n_loops = 100
        else:
            n_loops = 1

        # eofs / Dask
        try:
            if nbytes < 10_000:  # There seems to be a limit with eofs / Dask
                print("eofs / Dask ...")
                t_eofs = timeit.Timer(lambda: fit_eofs(X))
                timings["eofs_dask"][i][j] = (
                    np.array(t_eofs.repeat(n_repeats, n_loops)) / n_loops
                )
        except Exception as e:
            print(f"Warning: eofs Dask failed for {ns} samples / {nf} features")
            print("Error message: ", e)

        # xeofs / Dask
        try:
            print("xeofs / Dask ...")
            t_xeofs = timeit.Timer(lambda: fit_xeofs(X))
            timings["xeofs_dask"][i][j] = (
                np.array(t_xeofs.repeat(n_repeats, n_loops)) / n_loops
            )
        except Exception as e:
            print(f"Warning: xeofs Dask failed for {ns} samples / {nf} features")
            print("Error message: ", e)

        # eofs / no Dask
        if nbytes < MAX_MEMORY["eofs"]:
            try:
                print("eofs ...")
                X = X.load()
                t_eofs = timeit.Timer(lambda: fit_eofs(X))
                timings["eofs"][i][j] = (
                    np.array(t_eofs.repeat(n_repeats, n_loops)) / n_loops
                )
            except Exception as e:
                print(f"Warning: eofs failed for {ns} samples / {nf} features")
                print("Error message: ", e)

        # xeofs / no Dask
        if nbytes < MAX_MEMORY["xeofs"]:
            try:
                print("xeofs ...")
                X = X.load()
                t_xeofs = timeit.Timer(lambda: fit_xeofs(X))
                timings["xeofs"][i][j] = (
                    np.array(t_xeofs.repeat(n_repeats, n_loops)) / n_loops
                )
            except Exception as e:
                print(f"Warning: xeofs failed for {ns} samples / {nf} features")
                print("Error message: ", e)

# %%
# Save timings
da_size = xr.DataArray(
    timings["data"],
    dims=("n_samples", "n_features"),
    coords={"n_samples": n_samples, "n_features": n_features_x * 10},
    name="nbytes",
)


da1 = xr.DataArray(
    timings["eofs"],
    dims=("n_samples", "n_features", "run"),
    coords={
        "n_samples": n_samples,
        "n_features": n_features_x * 10,
        "run": np.arange(n_repeats),
    },
    name="time",
)
da2 = xr.DataArray(
    timings["xeofs"],
    dims=("n_samples", "n_features", "run"),
    coords={
        "n_samples": n_samples,
        "n_features": n_features_x * 10,
        "run": np.arange(n_repeats),
    },
    name="time",
)

da3 = xr.DataArray(
    timings["eofs_dask"],
    dims=("n_samples", "n_features", "run"),
    coords={
        "n_samples": n_samples,
        "n_features": n_features_x * 10,
        "run": np.arange(n_repeats),
    },
    name="time",
)
da4 = xr.DataArray(
    timings["xeofs_dask"],
    dims=("n_samples", "n_features", "run"),
    coords={
        "n_samples": n_samples,
        "n_features": n_features_x * 10,
        "run": np.arange(n_repeats),
    },
    name="time",
)

# %%
ds = xr.Dataset(
    {"nbytes": da_size, "eofs": da1, "xeofs": da2, "eofs_dask": da3, "xeofs_dask": da4}
)
ds.to_netcdf("timings.nc")


# %%



from typing import Any, Hashable, Sequence, TypeVar

import numpy as np
import xarray as xr

from .constants import VALID_LATITUDE_NAMES
from .data_types import (
    DaskArray,
    Data,
    DataArray,
    DataList,
    DataSet,
    DataVar,
    Dims,
    DimsList,
)
from .sanity_checks import convert_to_dim_type

T = TypeVar("T")


def unwrap_singleton_list(input_list: list[T]) -> T | list[T]:
    if len(input_list) == 1:
        return input_list[0]
    else:
        return input_list


def data_is_dask(data: DataArray | DataSet | DataList) -> bool:
    """Check if the given data is backed by a dask array."""

    # If data is a DataArray, check its underlying data type
    if isinstance(data, DataArray):
        return isinstance(data.data, DaskArray)

    # If data is a DataSet, recursively check all contained DataArrays
    if isinstance(data, DataSet):
        return any(data_is_dask(da) for da in data.data_vars.values())

    # If data is a list, recursively check each element in the list
    if isinstance(data, list):
        return any(data_is_dask(da) for da in data)

    # If none of the above, the data type is unrecognized
    raise ValueError("unrecognized data type.")


def process_parameter(
    parameter_name: str, parameter, default, n_data: int
) -> list[Any]:
    if parameter is None:
        return convert_to_list(default) * n_data
    elif isinstance(parameter, (list, tuple)):
        _check_parameter_number(parameter_name, parameter, n_data)
        return convert_to_list(parameter)
    else:
        return convert_to_list(parameter) * n_data


def convert_to_list(data: T | list[T] | tuple[T]) -> list[T]:
    if isinstance(data, list):
        return data
    elif isinstance(data, tuple):
        return list(data)
    else:
        return list([data])


def _check_parameter_number(parameter_name: str, parameter, n_data: int):
    if len(parameter) != n_data:
        raise ValueError(
            f"number of data objects passed should match number of parameter {parameter_name}"
            f"len(data objects)={n_data} and "
            f"len({parameter_name})={len(parameter)}"
        )


def feature_ones_like(data: DataVar, feature_dims: Dims) -> DataVar:
    if isinstance(data, xr.DataArray):
        valid_dims = set(data.dims) & set(feature_dims)
        feature_coords = {dim: data[dim] for dim in valid_dims}
        shape = tuple(coords.size for coords in feature_coords.values())
        return xr.DataArray(
            np.ones(shape, dtype=float),
            dims=tuple(valid_dims),
            coords=feature_coords,
        )
    elif isinstance(data, xr.Dataset):
        return xr.Dataset(
            {
                var: feature_ones_like(da, feature_dims)
                for var, da in data.data_vars.items()
            }
        )
    else:
        raise TypeError(
            "Invalid input type: {:}. Expected one of the following: DataArray or Dataset".format(
                type(data).__name__
            )
        )


def compute_sqrt_cos_lat_weights(data: DataVar, feature_dims: Dims) -> DataVar:
    """Compute the square root of cosine of latitude weights.

    Parameters
    ----------
    data : xr.DataArray | xr.Dataset
        Data to be scaled.
    dim : sequence of hashable
        Dimensions along which the data is considered to be a feature.

    Returns
    -------
    xr.DataArray | xr.Dataset
        Square root of cosine of latitude weights.

    """

    if isinstance(data, xr.DataArray):
        lat_dim = extract_latitude_dimension(feature_dims)

        latitudes = data.coords[lat_dim]
        weights = sqrt_cos_lat_weights(latitudes)
        # Features that cannot be associated to a latitude receive a weight of 1
        # weights = weights.where(weights.notnull(), 1)
        weights.name = "coslat_weights"
        return weights
    elif isinstance(data, xr.Dataset):
        return xr.Dataset(
            {
                var: compute_sqrt_cos_lat_weights(da, feature_dims)
                for var, da in data.data_vars.items()
            }
        )
    else:
        raise TypeError(
            "Invalid input type: {:}. Expected one of the following: DataArray".format(
                type(data).__name__
            )
        )


def extract_latitude_dimension(feature_dims: Dims) -> Hashable:
    # Find latitude coordinate
    lat_dim = set(feature_dims) & set(VALID_LATITUDE_NAMES)

    if len(lat_dim) == 0:
        raise ValueError(
            "No latitude coordinate was found to compute coslat weights. Must be one of the following: {:}".format(
                VALID_LATITUDE_NAMES
            )
        )
    elif len(lat_dim) == 1:
        return lat_dim.pop()
    else:
        raise ValueError(
            f"Found ambiguous latitude dimensions: {lat_dim}. Only ONE of the following is allowed for computing coslat weights: {VALID_LATITUDE_NAMES}"
        )


def get_dims(
    data: DataList,
    sample_dims: Hashable | Sequence[Hashable],
) -> tuple[Dims, DimsList]:
    """Extracts the dimensions of a DataArray or Dataset that are not included in the sample dimensions.

    Parameters:
    ------------
    data: xr.DataArray or xr.Dataset or list of xr.DataArray
        Input data.
    sample_dims: Hashable or Sequence[Hashable] or list[Sequence[Hashable]]
        Sample dimensions.

    Returns:
    ---------
    sample_dims: tuple[Hashable]
        Sample dimensions.
    feature_dims: tuple[Hashable]
        Feature dimensions.

    """
    # Check for invalid types
    if isinstance(data, list):
        sample_dims = convert_to_dim_type(sample_dims)
        feature_dims: DimsList = [_get_feature_dims(da, sample_dims) for da in data]
        return sample_dims, feature_dims
    else:
        err_message = f"Invalid input type: {type(data).__name__}. Expected one of "
        err_message += "of the following: list of DataArrays or Datasets."
        raise TypeError(err_message)


def _get_feature_dims(data: DataArray | DataSet, sample_dims: Dims) -> Dims:
    """Extracts the dimensions of a DataArray that are not included in the sample dimensions.


    Parameters:
    ------------
    data: xr.DataArray or xr.Dataset
        Input data.
    sample_dims: tuple[str]
        Sample dimensions.

    Returns:
    ---------
    feature_dims: tuple[str]
        Feature dimensions.

    """
    return tuple(dim for dim in data.dims if dim not in sample_dims)


def sqrt_cos_lat_weights(data: DataArray) -> DataArray:
    """Compute the square root of the cosine of the latitude.

    Parameters:
    ------------
    data: xr.DataArray
        Input data.

    Returns:
    ---------
    sqrt_cos_lat: xr.DataArray
        Square root of the cosine of the latitude.

    """
    return xr.apply_ufunc(
        _np_sqrt_cos_lat_weights,
        data,
        vectorize=False,
        dask="allowed",
    )


def total_variance(data: DataArray, dim) -> DataArray:
    """Compute the total variance of the input data.

    Parameters:
    ------------
    data: DataArray
        Input data.

    dim: str
        Dimension along which to compute the total variance.

    Returns:
    ---------
    tot_var: DataArray
        Total variance of the input data.

    """
    return data.var(dim, ddof=1).sum()


def _np_sqrt_cos_lat_weights(data):
    """Compute the square root of the cosine of the latitude.

    Parameters:
    ------------
    data: np.ndarray
        Input data.

    Returns:
    ---------
    sqrt_cos_lat: np.ndarray
        Square root of the cosine of the latitude.

    """
    return np.sqrt(np.cos(np.deg2rad(data)).clip(0, 1))


def get_deterministic_sign_multiplier(data: DataArray, dim: str) -> DataArray:
    """Compute a sign multiplier that ensures deterministic output.

    Uses a method standard to other SVD implementations where we ensure that
    the maximum absolute value in the passed data matrix is positive
    when multipled by the sign multiplier. This creates deterministic output.

    Parameters:
    ------------
    data: DataArray
        Input data to determine sorting order.
    dim: str
        Dimension along which to compute the sign multiplier.

    Returns:
    ---------
    sign_multiplier: DataArray
        Sign multiplier that ensures deterministic output.
    """
    # This method is carefully constructed to avoid idexing ops
    # so we can execute this lazily on dask arrays
    min_max = xr.concat([data.max(dim), data.min(dim)], dim="sign")
    min_max = min_max.assign_coords(sign=[1, -1])
    sign_multiplier = np.abs(min_max).idxmax("sign")
    # Drop all dimensions except 'mode' so that the index is clean
    for dim, coords in sign_multiplier.coords.items():
        if dim != "mode":
            sign_multiplier = sign_multiplier.drop(dim)
    return sign_multiplier


def argsort_dask(data: Data, dim: str) -> Data:
    """Apply argsort to a dask-backed array.

    This is a workaround because dask does not yet implement a chunk-aware
    version of argsort. Therefore we have to force rechunking to a single
    chunk along the sorting dimension and then apply numpy argsort. This
    should be used with an understanding that it may produce large memory
    usage if the passed array is chunked finely along the sorting dimension.

    Parameters:
    ------------
    data: Data
        Input data to sort.
    dim: str
        Dimension along which to sort.

    Returns:
    ---------
    sorted_idx: Data
        Indices that would sort the data.
    """
    return xr.apply_ufunc(
        np.argsort,
        data.chunk({dim: -1}),
        input_core_dims=[[dim]],
        output_core_dims=[[dim]],
        dask="parallelized",
    )


def get_matrix_rank(data: DataArray) -> int:
    """Compute the rank of a 2D matrix.

    We use the imprecice but practical assumption for real-world data
    that the rank of the matrix is the minimum of the number of rows and columns.

    Parameters:
    ------------
    data: DataArray
        Input data.

    Returns:
    ---------
    rank: int
        Rank of the input matrix.

    """
    if len(data.dims) != 2:
        raise ValueError("Input data must be a 2D matrix.")
    return min(data.shape)



from typing import Hashable, Sequence, TypeAlias, TypeVar

import dask.array as da
from xarray.core import dataarray as xr_dataarray
from xarray.core import dataset as xr_dataset

DataArray: TypeAlias = xr_dataarray.DataArray
DataSet: TypeAlias = xr_dataset.Dataset
Data: TypeAlias = DataArray | DataSet
DataVar = TypeVar("DataVar", DataArray, DataSet)
DataVarBound = TypeVar("DataVarBound", bound=Data)

DataArrayList: TypeAlias = list[DataArray]
DataSetList: TypeAlias = list[DataSet]
DataList: TypeAlias = list[Data]
DataVarList: TypeAlias = list[DataVar]

GenericType = TypeVar("GenericType")

DaskArray: TypeAlias = da.Array  # type: ignore
DataObject: TypeAlias = DataArray | DataSet | DataList

Dims: TypeAlias = Sequence[Hashable]
DimsTuple: TypeAlias = tuple[Dims, ...]
DimsList: TypeAlias = list[Dims]
DimsListTuple: TypeAlias = tuple[DimsList, ...]



import math

import numpy as np
import pytest
import xarray as xr

from xeofs.preprocessing import Whitener

from ..conftest import generate_synthetic_dataarray
from ..utilities import (
    assert_expected_coords,
    assert_expected_dims,
    data_is_dask,
)

# =============================================================================
# GENERALLY VALID TEST CASES
# =============================================================================
N_SAMPLE_DIMS = [1]
N_FEATURE_DIMS = [1]
INDEX_POLICY = ["index"]
NAN_POLICY = ["no_nan"]
DASK_POLICY = ["no_dask", "dask"]
SEED = [0]

VALID_TEST_DATA = [
    (ns, nf, index, nan, dask)
    for ns in N_SAMPLE_DIMS
    for nf in N_FEATURE_DIMS
    for index in INDEX_POLICY
    for nan in NAN_POLICY
    for dask in DASK_POLICY
]


def generate_well_conditioned_data(lazy=False):
    t = np.linspace(0, 50, 200)
    std = 0.1
    X = np.sin(t)[:, None] + np.random.normal(0, std, size=(200, 3))
    X[:, 1] = X[:, 1] ** 3
    X[:, 2] = abs(X[:, 2]) ** (0.5)
    X = xr.DataArray(
        X,
        dims=["sample", "feature"],
        coords={"sample": np.arange(200), "feature": np.arange(3)},
        name="X",
    )
    X = X - X.mean("sample")
    if lazy:
        X = X.chunk({"sample": 5, "feature": -1})
    return X


# TESTS
# =============================================================================
@pytest.mark.parametrize(
    "lazy",
    [False, True],
)
def test_fit(lazy):
    data = generate_well_conditioned_data(lazy)

    whitener = Whitener()
    whitener.fit(data)


@pytest.mark.parametrize("lazy", [False, True])
@pytest.mark.parametrize("alpha", [0.0, 0.5, 1.0, 1.5])
def test_transform(lazy, alpha):
    data = generate_well_conditioned_data(lazy)

    whitener = Whitener()
    whitener.fit(data)

    # Transform data
    transformed_data = whitener.transform(data)
    transformed_data2 = whitener.transform(data)

    assert isinstance(transformed_data, xr.DataArray)
    assert transformed_data.ndim == 2
    assert transformed_data.dims == ("sample", "feature")

    # Transformed data is identical
    assert transformed_data.identical(transformed_data2)

    # Check that for full whitening, transformed data is uncorrelated
    # and has variance one
    if math.isclose(alpha, 0.0, abs_tol=1e-6):
        # Transformed data has variance = 1
        assert np.allclose(transformed_data.var("sample").values, 1.0)

        # Transformed data is uncorrelated
        C = np.corrcoef(transformed_data.values, rowvar=False)
        target = np.identity(data.shape[1])
        np.testing.assert_allclose(C, target, atol=1e-6)

    # Consistent dask behaviour
    is_dask_before = data_is_dask(data)
    is_dask_after = data_is_dask(transformed_data)
    assert is_dask_before == is_dask_after


@pytest.mark.parametrize("lazy", [False, True])
def test_fit_transform(lazy):
    data = generate_well_conditioned_data(lazy)

    whitener = Whitener()

    # Transform data
    transformed_data = whitener.fit_transform(data)
    transformed_data2 = whitener.transform(data)
    assert transformed_data.identical(transformed_data2)

    assert isinstance(transformed_data, xr.DataArray)
    assert transformed_data.ndim == 2
    assert transformed_data.dims == ("sample", "feature")

    # Consistent dask behaviour
    is_dask_before = data_is_dask(data)
    is_dask_after = data_is_dask(transformed_data)
    assert is_dask_before == is_dask_after


@pytest.mark.parametrize("lazy", [False, True])
@pytest.mark.parametrize("alpha", [0.0, 0.5, 1.0, 1.5])
def test_inverse_transform_data(lazy, alpha):
    data = generate_well_conditioned_data(lazy)

    whitener = Whitener(alpha)
    whitener.fit(data)

    whitened_data = whitener.transform(data)
    unwhitened_data = whitener.inverse_transform_data(whitened_data)

    is_dask_before = data_is_dask(data)
    is_dask_after = data_is_dask(unwhitened_data)

    # Unstacked data has dimensions of original data
    assert_expected_dims(data, unwhitened_data, policy="all")
    # Unstacked data has coordinates of original data
    assert_expected_coords(data, unwhitened_data, policy="all")
    # inverse transform should not change dask-ness
    assert is_dask_before == is_dask_after

    # Unwhitened data is identical to original data
    xr.testing.assert_allclose(data, unwhitened_data, atol=1e-6)


def test_invalid_alpha():
    data = generate_synthetic_dataarray(1, 1, "index", "no_nan", "no_dask")
    data = data.rename({"sample0": "sample", "feature0": "feature"})

    err_msg = "`alpha` must be greater than or equal to 0"
    with pytest.raises(ValueError, match=err_msg):
        Whitener(alpha=-1.0)


def test_raise_warning_ill_conditioned():
    # Synthetic dataset has 6 samples and 7 features
    data = generate_synthetic_dataarray(1, 1, "index", "no_nan", "no_dask")
    data = data.rename({"sample0": "sample", "feature0": "feature"})

    whitener = Whitener()
    warn_msg = "The number of samples (.*) is smaller than the number of features (.*), leading to an ill-conditioned problem.*"
    with pytest.warns(match=warn_msg):
        _ = whitener.fit_transform(data)


def test_whitener_identity_transformation():
    data = generate_well_conditioned_data()

    whitener = Whitener(alpha=1.0)
    transformed = whitener.fit_transform(data)
    reconstructed = whitener.inverse_transform_data(transformed)

    # No transformation takes place
    xr.testing.assert_identical(data, transformed)
    xr.testing.assert_identical(data, reconstructed)


@pytest.mark.parametrize("alpha", [0.0, 0.5, 1.0, 1.5])
def test_unwhiten_cross_covariance_matrix(alpha):
    def unwhiten_cov_mat(Cw, S1_inv, S2_inv):
        n_dims_s1 = len(S1_inv.shape)
        n_dims_s2 = len(S2_inv.shape)

        match n_dims_s1:
            case 0:
                C_unwhitened = Cw
            case 2:
                C_unwhitened = S1_inv @ Cw
            case _:
                raise ValueError("Invalid number of dimensions for S1_inv")

        match n_dims_s2:
            case 0:
                pass
            case 2:
                C_unwhitened = C_unwhitened @ S2_inv
            case _:
                raise ValueError("Invalid number of dimensions for S2_inv")

        return C_unwhitened

    """Test that we can uncover the original total amount of squared covariance between two datasets after whitening."""
    data1 = generate_well_conditioned_data()
    data2 = generate_well_conditioned_data() ** 2

    whitener1 = Whitener(alpha=0.5)
    whitener2 = Whitener(alpha=alpha)

    transformed1 = whitener1.fit_transform(data1)
    transformed2 = whitener2.fit_transform(data2)

    S1_inv = whitener1.Tinv.values
    S2_inv = whitener2.Tinv.values

    C = data1.values.T @ data2.values
    Cw = transformed1.values.T @ transformed2.values

    C_rec = unwhiten_cov_mat(Cw, S1_inv, S2_inv)

    total_covariance = np.linalg.norm(C)
    total_covariance_rec = np.linalg.norm(C_rec)
    np.testing.assert_almost_equal(total_covariance, total_covariance_rec)


@pytest.mark.parametrize("alpha", [0.0, 0.5, 1.0, 1.5])
def test_transform_keep_coordinates(alpha):
    X = generate_well_conditioned_data()

    whitener = Whitener(alpha=alpha)
    transformed = whitener.fit_transform(X)

    assert len(transformed.coords) == len(X.coords)



import pytest

from xeofs.preprocessing import Stacker
from xeofs.utils.data_types import DataArray
from ..conftest import generate_synthetic_dataset
from ..utilities import (
    get_dims_from_data,
    data_is_dask,
    assert_expected_dims,
    assert_expected_coords,
)

# =============================================================================
# GENERALLY VALID TEST CASES
# =============================================================================
N_VARIABLES = [1, 2]
N_SAMPLE_DIMS = [1, 2]
N_FEATURE_DIMS = [1, 2]
INDEX_POLICY = ["index"]
NAN_POLICY = ["no_nan"]
DASK_POLICY = ["no_dask", "dask"]
SEED = [0]

VALID_TEST_DATA = [
    (nv, ns, nf, index, nan, dask)
    for nv in N_VARIABLES
    for ns in N_SAMPLE_DIMS
    for nf in N_FEATURE_DIMS
    for index in INDEX_POLICY
    for nan in NAN_POLICY
    for dask in DASK_POLICY
]


# TESTS
# =============================================================================
@pytest.mark.parametrize(
    "sample_name, feature_name, data_params",
    [
        ("sample", "feature", (1, 1, 1)),
        ("sample", "feature", (2, 1, 1)),
        ("sample", "feature", (1, 2, 2)),
        ("sample", "feature", (2, 2, 2)),
        ("sample0", "feature", (1, 1, 1)),
        ("sample0", "feature", (1, 1, 2)),
        ("sample0", "feature", (2, 1, 2)),
        ("another_sample", "another_feature", (1, 1, 1)),
        ("another_sample", "another_feature", (1, 2, 2)),
        ("another_sample", "another_feature", (2, 1, 1)),
        ("another_sample", "another_feature", (2, 2, 2)),
    ],
)
def test_fit_valid_dimension_names(sample_name, feature_name, data_params):
    data = generate_synthetic_dataset(*data_params)
    all_dims, sample_dims, feature_dims = get_dims_from_data(data)

    stacker = Stacker(sample_name=sample_name, feature_name=feature_name)
    stacker.fit(data, sample_dims, feature_dims)
    stacked_data = stacker.transform(data)
    reconstructed_data = stacker.inverse_transform_data(stacked_data)

    assert stacked_data.ndim == 2
    assert set(stacked_data.dims) == set((sample_name, feature_name))
    assert set(reconstructed_data.dims) == set(data.dims)


@pytest.mark.parametrize(
    "sample_name, feature_name, data_params",
    [
        ("sample", "feature0", (1, 1, 1)),
        ("sample0", "feature", (1, 2, 1)),
        ("sample1", "feature1", (1, 3, 3)),
        ("sample", "feature0", (2, 1, 1)),
        ("sample0", "feature", (2, 2, 1)),
        ("sample1", "feature1", (2, 3, 3)),
    ],
)
def test_fit_invalid_dimension_names(sample_name, feature_name, data_params):
    data = generate_synthetic_dataset(*data_params)
    all_dims, sample_dims, feature_dims = get_dims_from_data(data)

    stacker = Stacker(sample_name=sample_name, feature_name=feature_name)

    with pytest.raises(ValueError):
        stacker.fit(data, sample_dims, feature_dims)


@pytest.mark.parametrize(
    "synthetic_dataset",
    VALID_TEST_DATA,
    indirect=["synthetic_dataset"],
)
def test_fit(synthetic_dataset):
    data = synthetic_dataset
    all_dims, sample_dims, feature_dims = get_dims_from_data(data)

    stacker = Stacker()
    stacker.fit(data, sample_dims, feature_dims)


@pytest.mark.parametrize(
    "synthetic_dataset",
    VALID_TEST_DATA,
    indirect=["synthetic_dataset"],
)
def test_transform(synthetic_dataset):
    data = synthetic_dataset
    all_dims, sample_dims, feature_dims = get_dims_from_data(data)

    stacker = Stacker()
    stacker.fit(data, sample_dims, feature_dims)
    transformed_data = stacker.transform(data)
    transformed_data2 = stacker.transform(data)

    is_dask_before = data_is_dask(data)
    is_dask_after = data_is_dask(transformed_data)

    assert isinstance(transformed_data, DataArray)
    assert transformed_data.ndim == 2
    assert transformed_data.dims == ("sample", "feature")
    assert is_dask_before == is_dask_after
    assert transformed_data.identical(transformed_data2)


@pytest.mark.parametrize(
    "synthetic_dataset",
    VALID_TEST_DATA,
    indirect=["synthetic_dataset"],
)
def test_transform_invalid(synthetic_dataset):
    data = synthetic_dataset
    all_dims, sample_dims, feature_dims = get_dims_from_data(data)

    stacker = Stacker()
    stacker.fit(data, sample_dims, feature_dims)
    with pytest.raises(ValueError):
        stacker.transform(data.isel(feature0=slice(0, 2)))


@pytest.mark.parametrize(
    "synthetic_dataset",
    VALID_TEST_DATA,
    indirect=["synthetic_dataset"],
)
def test_fit_transform(synthetic_dataset):
    data = synthetic_dataset
    all_dims, sample_dims, feature_dims = get_dims_from_data(data)

    stacker = Stacker()
    transformed_data = stacker.fit_transform(data, sample_dims, feature_dims)

    is_dask_before = data_is_dask(data)
    is_dask_after = data_is_dask(transformed_data)

    assert isinstance(transformed_data, DataArray)
    assert transformed_data.ndim == 2
    assert transformed_data.dims == ("sample", "feature")
    assert is_dask_before == is_dask_after


@pytest.mark.parametrize(
    "synthetic_dataset",
    VALID_TEST_DATA,
    indirect=["synthetic_dataset"],
)
def test_invserse_transform_data(synthetic_dataset):
    data = synthetic_dataset
    all_dims, sample_dims, feature_dims = get_dims_from_data(data)

    stacker = Stacker()
    stacker.fit(data, sample_dims, feature_dims)
    stacked_data = stacker.transform(data)
    unstacked_data = stacker.inverse_transform_data(stacked_data)

    is_dask_before = data_is_dask(data)
    is_dask_after = data_is_dask(unstacked_data)

    # Unstacked data has dimensions of original data
    assert_expected_dims(data, unstacked_data, policy="all")
    # Unstacked data has coordinates of original data
    assert_expected_coords(data, unstacked_data, policy="all")
    # inverse transform should not change dask-ness
    assert is_dask_before == is_dask_after


@pytest.mark.parametrize(
    "synthetic_dataset",
    VALID_TEST_DATA,
    indirect=["synthetic_dataset"],
)
def test_invserse_transform_components(synthetic_dataset):
    data = synthetic_dataset
    all_dims, sample_dims, feature_dims = get_dims_from_data(data)

    stacker = Stacker()
    stacker.fit(data, sample_dims, feature_dims)

    stacked_data = stacker.transform(data)
    components = stacked_data.rename({"sample": "mode"})
    components.coords.update({"mode": range(components.mode.size)})
    unstacked_data = stacker.inverse_transform_components(components)

    is_dask_before = data_is_dask(data)
    is_dask_after = data_is_dask(unstacked_data)

    # Unstacked components has correct feature dimensions
    assert_expected_dims(data, unstacked_data, policy="feature")
    # Unstacked data has coordinates of original data
    assert_expected_coords(data, unstacked_data, policy="feature")
    # inverse transform should not change dask-ness
    assert is_dask_before == is_dask_after


@pytest.mark.parametrize(
    "synthetic_dataset",
    VALID_TEST_DATA,
    indirect=["synthetic_dataset"],
)
def test_invserse_transform_scores(synthetic_dataset):
    data = synthetic_dataset
    all_dims, sample_dims, feature_dims = get_dims_from_data(data)

    stacker = Stacker()
    stacker.fit(data, sample_dims, feature_dims)

    stacked_data = stacker.transform(data)
    scores = stacked_data.rename({"feature": "mode"})
    scores.coords.update({"mode": range(scores.mode.size)})
    unstacked_data = stacker.inverse_transform_scores(scores)

    is_dask_before = data_is_dask(data)
    is_dask_after = data_is_dask(unstacked_data)

    # Unstacked scores has correct feature dimensions
    assert_expected_dims(data, unstacked_data, policy="sample")
    # Unstacked data has coordinates of original data
    assert_expected_coords(data, unstacked_data, policy="sample")
    # inverse transform should not change dask-ness
    assert is_dask_before == is_dask_after



import xarray as xr
from dask.base import compute
from typing_extensions import Self

from ..utils.data_types import Data, DataArray, Dims
from ..utils.sanity_checks import assert_single_dataarray
from .transformer import Transformer


class Sanitizer(Transformer):
    """
    Removes NaNs from the feature dimension of a 2D DataArray.

    """

    def __init__(self, sample_name="sample", feature_name="feature", check_nans=True):
        super().__init__(sample_name=sample_name, feature_name=feature_name)

        self.check_nans = check_nans

        self.feature_coords = xr.DataArray()
        self.sample_coords = xr.DataArray()
        self.is_valid_feature = xr.DataArray()

    def get_serialization_attrs(self) -> dict:
        return dict(
            feature_coords=self.feature_coords,
            sample_coords=self.sample_coords,
            is_valid_feature=self.is_valid_feature,
        )

    def _check_input_dims(self, X) -> None:
        if set(X.dims) != set([self.sample_name, self.feature_name]):
            raise ValueError(
                "Input must have dimensions ({:}, {:})".format(
                    self.sample_name, self.feature_name
                )
            )

    def _check_input_coords(self, X) -> None:
        if not X.coords[self.feature_name].identical(self.feature_coords):
            raise ValueError(
                "Cannot transform data. Feature coordinates are different."
            )

    def _get_valid_features(self, X: Data) -> Data:
        return X.notnull().any(self.sample_name)

    def _get_valid_samples(self, X: Data) -> Data:
        return X.notnull().any(self.feature_name)

    def _get_valid_features_per_sample(self, X: Data) -> Data:
        """Isolated NaN check that an be constructed lazily, where we check that
        the number of valid features is the same for every sample, with
        the exception of all-NaN samples."""
        return X.notnull().sum(self.feature_name)

    def fit(
        self,
        X: Data,
        sample_dims: Dims | None = None,
        feature_dims: Dims | None = None,
        **kwargs,
    ) -> Self:
        # Check if input is a DataArray
        assert_single_dataarray(X)

        # Check if input has the correct dimensions
        self._check_input_dims(X)

        self.feature_coords = X.coords[self.feature_name]
        self.sample_coords = X.coords[self.sample_name]

        # Identify NaN locations
        self.is_valid_feature = self._get_valid_features(X)
        self.is_valid_sample = self._get_valid_samples(X)

        return self

    def transform(self, X: DataArray) -> DataArray:
        # Check if input is a DataArray
        assert_single_dataarray(X)

        # Check if input has the correct dimensions
        self._check_input_dims(X)

        # Check if input has the correct coordinates
        self._check_input_coords(X)

        X_valid_features = self._get_valid_features(X)
        X_valid_samples = self._get_valid_samples(X)
        X_valid_features_per_sample = self._get_valid_features_per_sample(X)

        # Optionally skip NaN checks to preserve lazy computation for dask arrays
        if self.check_nans:
            (
                self.is_valid_feature,
                X_valid_features,
                X_valid_samples,
                X_valid_features_per_sample,
            ) = compute(
                self.is_valid_feature,
                X_valid_features,
                X_valid_samples,
                X_valid_features_per_sample,
            )

            # Validate that non-NaN features match the original from .fit()
            if not X_valid_features.equals(self.is_valid_feature):
                raise ValueError(
                    "Input data had NaN features in different locations than"
                    " the original data."
                )

            isolated_nans = ~X_valid_features_per_sample.isin(
                [0, X_valid_features.sum().values]
            )
            if isolated_nans.any():
                raise ValueError(
                    "Input data contains partial NaN entries, which will cause the"
                    " the SVD to fail."
                )

            X = X.where(X_valid_features & X_valid_samples, drop=True)

        return X

    def inverse_transform_data(self, X: DataArray) -> DataArray:
        # Reindex only if feature coordinates are different
        coords_are_equal = X.coords[self.feature_name].identical(self.feature_coords)

        if coords_are_equal:
            return X
        else:
            return X.reindex({self.feature_name: self.feature_coords.values})

    def inverse_transform_components(self, X: DataArray) -> DataArray:
        # Reindex only if feature coordinates are different
        coords_are_equal = X.coords[self.feature_name].identical(self.feature_coords)

        if coords_are_equal:
            return X
        else:
            return X.reindex({self.feature_name: self.feature_coords.values})

    def inverse_transform_scores(self, X: DataArray) -> DataArray:
        # Reindex only if sample coordinates are different
        coords_are_equal = X.coords[self.sample_name].identical(self.sample_coords)

        if coords_are_equal:
            return X
        else:
            return X.reindex({self.sample_name: self.sample_coords.values})

    def inverse_transform_scores_unseen(self, X: DataArray) -> DataArray:
        # Don't check sample coords for unseen data
        return X



from abc import abstractmethod
from typing import Any, Hashable, Sequence

from numpy.random import Generator
from typing_extensions import Self

from ..base_model import BaseModel
from ..data_container import DataContainer
from ..preprocessing import PCA, Preprocessor, Whitener
from ..utils.data_types import DataArray, DataObject, GenericType
from ..utils.sanity_checks import validate_input_type
from ..utils.xarray_utils import convert_to_dim_type


class BaseModelCrossSet(BaseModel):
    """
    Abstract base class for cross-decomposition models.

    Parameters:
    -------------
    n_modes: int
        Number of modes to calculate.
    center: bool, default=True
        Whether to center the input data.
    standardize: bool, default=False
        Whether to standardize the input data.
    use_coslat: bool, default=False
        Whether to use cosine of latitude for scaling.
    check_nans : bool, default=True
        If True, remove full-dimensional NaN features from the data, check to
        ensure that NaN features match the original fit data during transform,
        and check for isolated NaNs. Note: this forces eager computation of dask
        arrays. If False, skip all NaN checks. In this case, NaNs should be
        explicitly removed or filled prior to fitting, or SVD will fail.
    alpha : float, default=1.0
        Parameter to perform fractional whitening of the data. If 0, the data is
        completely whitened. If 1, the data is not whitened.
    use_pca : bool, default=False
        If True, perform PCA to reduce the dimensionality of the data.
    n_pca_modes : int | float | str, default=0.999
        If int, specifies the number of modes to retain. If float, specifies the
        fraction of variance in the (whitened) data that should be explained by
        the retained modes. If "all", all modes are retained.
    init_rank_reduction : float, default=0.3
        Only relevant when `use_pca=True` and `n_modes` is a float, in which
        case it denotes the fraction of the initial rank to reduce the data to
        via PCA as a first guess before truncating the solution to the desired
        fraction of explained variance. This allows for faster computation of
        PCA via randomized SVD and avoids the need to compute the full SVD.
    compute: bool, default=True
        Whether to compute elements of the model eagerly, or to defer
        computation. If True, four pieces of the fit will be computed
        sequentially: 1) the preprocessor scaler, 2) optional NaN checks, 3) SVD
        decomposition, 4) scores and components.
    random_state: numpy.random.Generator or int, optional
        Seed for the random number generator.
    sample_name: str, default="sample"
        Name of the new sample dimension.
    feature_name: str, default="feature"
        Name of the new feature dimension.
    solver: {"auto", "full", "randomized"}
        Solver to use for the SVD computation.
    solver_kwargs: dict[str, Any], default={}
        Additional keyword arguments to passed to the SVD solver function.

    """

    def __init__(
        self,
        n_modes: int,
        center: Sequence[bool] | bool = True,
        standardize: Sequence[bool] | bool = False,
        use_coslat: Sequence[bool] | bool = False,
        check_nans: Sequence[bool] | bool = True,
        use_pca: Sequence[bool] | bool = True,
        n_pca_modes: Sequence[float | int | str] | float | int | str = 0.999,
        pca_init_rank_reduction: Sequence[float] | float = 0.3,
        alpha: Sequence[float] | float = 1.0,
        solver: Sequence[str] | str = "auto",
        compute: bool = True,
        sample_name: str = "sample",
        feature_name: Sequence[str] | str = "feature",
        random_state: Generator | int | None = None,
        solver_kwargs: dict[str, Any] = {},
    ):
        super().__init__()

        # Process parameters
        center = self._process_parameter("center", center, True)
        standardize = self._process_parameter("standardize", standardize, False)
        use_coslat = self._process_parameter("use_coslat", use_coslat, False)
        check_nans = self._process_parameter("check_nans", check_nans, True)
        use_pca = self._process_parameter("use_pca", use_pca, True)
        n_pca_modes = self._process_parameter("n_pca_modes", n_pca_modes, 0.999)
        pca_init_rank_reduction = self._process_parameter(
            "pca_init_rank_reduction", pca_init_rank_reduction, 0.3
        )

        alpha = self._process_parameter("alpha", alpha, 1.0)
        # Ensure that alpha is a float
        alpha = [float(a) for a in alpha]

        # Use feature1 and feature2 throughout the model to refer to the two datasets
        if isinstance(feature_name, str):
            feature_name = [feature_name + str(i + 1) for i in range(2)]
        self._check_parameter_number("feature_name", feature_name)
        if feature_name[0] == feature_name[1]:
            raise ValueError("feature_name must be different for each dataset")

        # Define model parameters
        self.sample_name = sample_name
        self.feature_name = feature_name

        self._params = {
            "n_modes": n_modes,
            "center": center,
            "standardize": standardize,
            "use_coslat": use_coslat,
            "check_nans": check_nans,
            "use_pca": use_pca,
            "n_pca_modes": n_pca_modes,
            "pca_init_rank_reduction": pca_init_rank_reduction,
            "alpha": alpha,
            "sample_name": sample_name,
            "feature_name": feature_name,
            "random_state": random_state,
            "compute": compute,
            "solver": solver,
        }

        self._decomposer_kwargs = {
            "n_modes": n_modes,
            "init_rank_reduction": pca_init_rank_reduction,
            "solver": solver,
            "random_state": random_state,
            "compute": compute,
            "solver_kwargs": solver_kwargs,
        }

        # Define analysis-relevant meta data
        self.attrs.update({"model": "BaseModelCrossSet"})
        self.attrs.update(self.get_params())

        # Initialize preprocessors for dataset X and Y
        self.preprocessor1 = Preprocessor(
            sample_name=sample_name,
            feature_name=feature_name[0],
            with_center=center[0],
            with_std=standardize[0],
            with_coslat=use_coslat[0],
            check_nans=check_nans[0],
            compute=compute,
        )

        self.preprocessor2 = Preprocessor(
            sample_name=sample_name,
            feature_name=feature_name[1],
            with_center=center[1],
            with_std=standardize[1],
            with_coslat=use_coslat[1],
            check_nans=check_nans[1],
            compute=compute,
        )

        self.pca1 = PCA(
            n_modes=n_pca_modes[0],
            init_rank_reduction=pca_init_rank_reduction[0],
            use_pca=use_pca[0],
            sample_name=sample_name,
            feature_name=feature_name[0],
        )

        self.pca2 = PCA(
            n_modes=n_pca_modes[1],
            init_rank_reduction=pca_init_rank_reduction[1],
            use_pca=use_pca[1],
            sample_name=sample_name,
            feature_name=feature_name[1],
        )

        self.whitener1 = Whitener(
            alpha=alpha[0],
            sample_name=sample_name,
            feature_name=feature_name[0],
        )
        self.whitener2 = Whitener(
            alpha=alpha[1],
            sample_name=sample_name,
            feature_name=feature_name[1],
        )

        # Initialize the data container that stores the results
        self.data = DataContainer()

    @abstractmethod
    def _fit_algorithm(self, X: DataArray, Y: DataArray) -> Self:
        """
        Fit the model to the preprocessed data. This method needs to be implemented in the respective
        subclass.

        Parameters
        ----------
        X, Y: DataArray
            Preprocessed input data of two dimensions: (`sample_name`, `feature_name`)

        """
        raise NotImplementedError

    @abstractmethod
    def _transform_algorithm(
        self, X: DataArray | None = None, Y: DataArray | None = None, **kwargs
    ) -> dict[str, DataArray]:
        """
        Transform the preprocessed data. This method needs to be implemented in the respective
        subclass.

        Parameters
        ----------
        X, Y: DataArray
            Preprocessed input data of two dimensions: (`sample_name`, `feature_name`)

        """
        raise NotImplementedError

    @abstractmethod
    def _inverse_transform_algorithm(
        self, X: DataArray | None = None, Y: DataArray | None = None, **kwargs
    ) -> dict[str, DataArray]:
        """
        Reconstruct the original data from transformed data. This method needs to be implemented in the respective
        subclass.

        Parameters
        ----------
        scores1: DataArray
            Transformed left field data to be reconstructed. This could be
            a subset of the `scores` data of a fitted model, or unseen data.
            Must have a 'mode' dimension.
        scores2: DataArray
            Transformed right field data to be reconstructed. This could be
            a subset of the `scores` data of a fitted model, or unseen data.
            Must have a 'mode' dimension.

        Returns
        -------
        Xrec1: DataArray
            Reconstructed data of left field.
        Xrec2: DataArray
            Reconstructed data of right field.

        """
        raise NotImplementedError

    @abstractmethod
    def _predict_algorithm(self, X: DataArray, **kwargs) -> DataArray:
        """Predict the right field from the left field. This method needs to be implemented in the respective subclass."""
        raise NotImplementedError

    @abstractmethod
    def _get_components(self, **kwargs) -> tuple[DataArray, DataArray]:
        """Get the components."""
        raise NotImplementedError

    @abstractmethod
    def _get_scores(self, **kwargs) -> tuple[DataArray, DataArray]:
        """Get the scores."""
        raise NotImplementedError

    def fit(
        self,
        X: DataObject,
        Y: DataObject,
        dim: Hashable | Sequence[Hashable],
        weights_X: DataObject | None = None,
        weights_Y: DataObject | None = None,
    ) -> Self:
        """Fit the data to the model.

        Parameters
        ----------
        X, Y: DataObject
            Data to be fitted.
        dim: Hashable | Sequence[Hashable]
            Define the sample dimensions. The remaining dimensions
            will be treated as feature dimensions.
        weights_X, weights_Y: DataObject | None, default=None
            Weights for the data. If None, no weights are used.

        Returns
        -------
        xeofs MultiSetModel
            Fitted model.

        """
        validate_input_type(X)
        validate_input_type(Y)
        if weights_X is not None:
            validate_input_type(weights_X)
        if weights_Y is not None:
            validate_input_type(weights_Y)

        self.sample_dims = convert_to_dim_type(dim)
        # Preprocess data
        X = self.preprocessor1.fit_transform(X, self.sample_dims, weights_X)
        Y = self.preprocessor2.fit_transform(Y, self.sample_dims, weights_Y)
        # Perform PCA
        X = self.pca1.fit_transform(X)
        Y = self.pca2.fit_transform(Y)
        # Augment data
        X, Y = self._augment_data(X, Y)
        # Whiten data
        X = self.whitener1.fit_transform(X)
        Y = self.whitener2.fit_transform(Y)
        # Fit the model
        self._fit_algorithm(X, Y)

        if self.get_params()["compute"]:
            self.data.compute()
            self._post_compute()

        return self

    def transform(
        self, X: DataObject | None = None, Y: DataObject | None = None, normalized=False
    ) -> Sequence[DataArray] | DataArray:
        """Transform the data.

        Parameters
        ----------
        X, Y: DataObject | None
            Data to be transformed. At least one of them must be provided.
        normalized: bool, default=False
            Whether to return L2 normalized scores.

        Returns
        -------
        Sequence[DataArray] | DataArray
            Transformed data.


        """
        if X is None and Y is None:
            raise ValueError("Either X or Y must be provided.")

        if X is not None:
            validate_input_type(X)
            # Preprocess X
            X = self.preprocessor1.transform(X)
            X = self.pca1.transform(X)
            X = self.whitener1.transform(X)
        if Y is not None:
            validate_input_type(Y)
            # Preprocess Y
            Y = self.preprocessor2.transform(Y)
            Y = self.pca2.transform(Y)
            Y = self.whitener2.transform(Y)

        data = self._transform_algorithm(X, Y, normalized=normalized)
        data_list = []
        if X is not None:
            X = self.whitener1.inverse_transform_scores_unseen(data["X"])
            X = self.pca1.inverse_transform_scores_unseen(X)
            X = self.preprocessor1.inverse_transform_scores_unseen(X)
            data_list.append(X)
        if Y is not None:
            Y = self.whitener2.inverse_transform_scores_unseen(data["Y"])
            Y = self.pca2.inverse_transform_scores_unseen(Y)
            Y = self.preprocessor2.inverse_transform_scores_unseen(Y)
            data_list.append(Y)

        if len(data_list) == 1:
            return data_list[0]
        else:
            return data_list

    def inverse_transform(
        self, X: DataArray | None = None, Y: DataArray | None = None
    ) -> Sequence[DataObject] | DataObject:
        """Reconstruct the original data from transformed data.

        Parameters
        ----------
        X, Y: DataArray | None
            Transformed data to be reconstructed. At least one of them must be provided.

        Returns
        -------
        Sequence[DataObject] | DataObject
            Reconstructed data.

        """
        x_is_given = X is not None
        y_is_given = Y is not None

        if (not x_is_given) and (not y_is_given):
            raise ValueError("Either X or Y must be provided.")

        if x_is_given:
            # Handle scalar mode in xr.dot
            if "mode" not in X.dims:
                X = X.expand_dims("mode")

        if y_is_given:
            # Handle scalar mode in xr.dot
            if "mode" not in Y.dims:
                Y = Y.expand_dims("mode")

        inv_transformed = self._inverse_transform_algorithm(X, Y)

        results: list[DataObject] = []

        # Unstack and rescale the data
        if x_is_given:
            X = inv_transformed["X"]
            X = self.whitener1.inverse_transform_data(X)
            X = self.pca1.inverse_transform_data(X)
            Xrec = self.preprocessor1.inverse_transform_data(X)
            results.append(Xrec)
        if y_is_given:
            Y = inv_transformed["Y"]
            Y = self.whitener2.inverse_transform_data(Y)
            Y = self.pca2.inverse_transform_data(Y)
            Yrec = self.preprocessor2.inverse_transform_data(Y)
            results.append(Yrec)

        if len(results) == 1:
            return results[0]
        else:
            return results

    def predict(self, X: DataObject) -> DataArray:
        """Predict Y from X.

        Parameters
        ----------
        X: DataObject
            Data to be used for prediction.

        Returns
        -------
        DataArray
            Predicted data in transformed space.

        """

        validate_input_type(X)

        # Preprocess X
        X = self.preprocessor1.transform(X)
        X = self.pca1.transform(X)

        # Whiten X
        X = self.whitener1.transform(X)

        # Predict Y
        Y = self._predict_algorithm(X)

        # Inverse transform Y
        Y = self.whitener2.inverse_transform_scores_unseen(Y)
        Y = self.pca2.inverse_transform_scores_unseen(Y)
        Y = self.preprocessor2.inverse_transform_scores_unseen(Y)

        return Y

    def components(self, normalized=True) -> tuple[DataObject, DataObject]:
        """Get the components of the model.

        The components may be referred to differently depending on the model
        type. Common terms include canonical vectors, singular vectors, loadings
        or spatial patterns.

        Parameters
        ----------
        normalized: bool, default=True
            Whether to return L2 normalized components.

        Returns
        -------
        tuple[DataObject, DataObject]
            Components of X and Y.

        """
        Px, Py = self._get_components(normalized=normalized)

        Px = self.whitener1.inverse_transform_components(Px)
        Py = self.whitener2.inverse_transform_components(Py)

        Px = self.pca1.inverse_transform_components(Px)
        Py = self.pca2.inverse_transform_components(Py)

        Px: DataObject = self.preprocessor1.inverse_transform_components(Px)
        Py: DataObject = self.preprocessor2.inverse_transform_components(Py)
        return Px, Py

    def scores(self, normalized=False) -> tuple[DataArray, DataArray]:
        """Get the scores of the model.

        The component scores may be referred to differently depending on the
        model type. Common terms include canonical variates, expansion
        coefficents, principal component (scores) or temporal patterns.

        Parameters
        ----------
        normalized: bool, default=False
            Whether to return L2 normalized scores.

        Returns
        -------
        tuple[DataArray, DataArray]
            Scores of X and Y.

        """
        Rx, Ry = self._get_scores(normalized=normalized)

        Rx = self.whitener1.inverse_transform_scores(Rx)
        Ry = self.whitener2.inverse_transform_scores(Ry)

        Rx = self.pca1.inverse_transform_scores(Rx)
        Ry = self.pca2.inverse_transform_scores(Ry)

        Rx: DataArray = self.preprocessor1.inverse_transform_scores(Rx)
        Ry: DataArray = self.preprocessor2.inverse_transform_scores(Ry)
        return Rx, Ry

    def get_serialization_attrs(self) -> dict:
        """Get the attributes needed to serialize the model.

        Returns
        -------
        dict
            Attributes needed to serialize the model.

        """
        return dict(
            data=self.data,
            preprocessor1=self.preprocessor1,
            preprocessor2=self.preprocessor2,
            pca1=self.pca1,
            pca2=self.pca2,
            whitener1=self.whitener1,
            whitener2=self.whitener2,
            sample_name=self.sample_name,
            feature_name=self.feature_name,
        )

    def _augment_data(self, X: DataArray, Y: DataArray) -> tuple[DataArray, DataArray]:
        """Optional method to augment the data before fitting."""
        return X, Y

    def _process_parameter(
        self,
        parameter_name: str,
        parameter: Sequence[GenericType] | GenericType | None,
        default: GenericType,
    ) -> Sequence[GenericType]:
        n_datasets = 2
        if parameter is None:
            parameter = [default] * n_datasets
        elif not isinstance(parameter, (list, tuple)):
            parameter = [parameter] * n_datasets
        self._check_parameter_number(parameter_name, parameter)
        return parameter

    @staticmethod
    def _check_parameter_number(
        parameter_name: str, parameter: Sequence[GenericType]
    ) -> None:
        if len(parameter) != 2:
            err_msg = (
                f"Expected 2 items for '{parameter_name}', but got {len(parameter)}."
            )
            raise ValueError(err_msg)



import xarray as xr

from ..utils.data_types import DataArray
from ._numpy._rotation import _promax


def promax(loadings: DataArray, feature_dim, **kwargs):
    rotated, rot_mat, phi_mat = xr.apply_ufunc(
        _promax,
        loadings,
        input_core_dims=[[feature_dim, "mode"]],
        output_core_dims=[
            [feature_dim, "mode"],
            ["mode_m", "mode_n"],
            ["mode_m", "mode_n"],
        ],
        kwargs=kwargs,
        dask="allowed",
    )

    return rotated, rot_mat, phi_mat



"""
Rotated Maximum Covariance Analysis
===================================

Rotated Maximum Covariance Analysis (MCA) between two data sets.
"""

# Load packages and data:
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from cartopy.crs import Orthographic, PlateCarree
from cartopy.feature import LAND
from matplotlib.gridspec import GridSpec

import xeofs as xe

# %%
# Create 2 different DataArrays

t2m = xr.tutorial.load_dataset("air_temperature")["air"]
da1 = t2m.isel(lon=slice(0, 26))
da2 = t2m.isel(lon=slice(27, None))

# %%
# Perform MCA

mca = xe.cross.MCA(n_modes=20, standardize=False, use_coslat=True)
mca.fit(da1, da2, dim="time")

# %%
# Apply Varimax-rotation to MCA solution

rot = xe.cross.MCARotator(n_modes=10)
rot.fit(mca)

# %%
# Get rotated singular vectors, projections (PCs), homogeneous and heterogeneous
# patterns:

singular_vectors = rot.components()
scores = rot.scores()
hom_pats, pvals_hom = rot.homogeneous_patterns()
het_pats, pvals_het = rot.heterogeneous_patterns()

# %%
# When two fields are expected, the output of the above methods is a list of
# length 2, with the first and second entry containing the relevant object for
# ``X`` and ``Y``. For example, the p-values obtained from the two-sided t-test
# for the homogeneous patterns of ``X`` are:

pvals_hom[0]

# %%
# Create a mask to identifiy where p-values are below 0.05

hom_mask = [values < 0.05 for values in pvals_hom]
het_mask = [values < 0.05 for values in pvals_het]


# %%
# Plot some relevant quantities of mode 2.

lonlats = [
    np.meshgrid(pvals_hom[0].lon.values, pvals_hom[0].lat.values),
    np.meshgrid(pvals_hom[1].lon.values, pvals_hom[1].lat.values),
]
proj = [
    Orthographic(central_latitude=30, central_longitude=-120),
    Orthographic(central_latitude=30, central_longitude=-60),
]
kwargs1 = {"cmap": "BrBG", "vmin": -0.05, "vmax": 0.05, "transform": PlateCarree()}
kwargs2 = {"cmap": "RdBu", "vmin": -1, "vmax": 1, "transform": PlateCarree()}

mode = 2

fig = plt.figure(figsize=(7, 14))
gs = GridSpec(5, 2)
ax1 = [fig.add_subplot(gs[0, i], projection=proj[i]) for i in range(2)]
ax2 = [fig.add_subplot(gs[1, i], projection=proj[i]) for i in range(2)]
ax3 = [fig.add_subplot(gs[2, i], projection=proj[i]) for i in range(2)]
ax4 = [fig.add_subplot(gs[3, i]) for i in range(2)]

for i, a in enumerate(ax1):
    singular_vectors[i].sel(mode=mode).plot(ax=a, **kwargs1)

for i, a in enumerate(ax2):
    hom_pats[i].sel(mode=mode).plot(ax=a, **kwargs2)
    a.scatter(
        lonlats[i][0],
        lonlats[i][1],
        hom_mask[i].sel(mode=mode).values * 0.5,
        color="k",
        alpha=0.5,
        transform=PlateCarree(),
    )
for i, a in enumerate(ax3):
    het_pats[i].sel(mode=mode).plot(ax=a, **kwargs2)
    a.scatter(
        lonlats[i][0],
        lonlats[i][1],
        het_mask[i].sel(mode=mode).values * 0.5,
        color="k",
        alpha=0.5,
        transform=PlateCarree(),
    )

for i, a in enumerate(ax4):
    scores[i].sel(mode=mode).plot(ax=a)
    a.set_xlabel("")


for a in np.ravel([ax1, ax2, ax3]):
    a.coastlines(color=".5")
    a.add_feature(LAND)

plt.tight_layout()
plt.savefig("rotated_mca.jpg")



import warnings

import numpy as np
from dask.array import Array as DaskArray  # type: ignore
from dask.array.linalg import svd_compressed as dask_svd
from dask.graph_manipulation import wait_on
from scipy.sparse.linalg import svds as complex_svd  # type: ignore
from sklearn.utils.extmath import randomized_svd

from ...utils.sanity_checks import sanity_check_n_modes


def get_deterministic_sign_multiplier(data, axis: int):
    """Compute a sign multiplier that ensures deterministic output using Dask.

    Parameters:
    ------------
    data: da.Array
        Input data to determine sorting order.
    axis: int
        Axis along which to compute the sign multiplier.

    Returns:
    ---------
    sign_multiplier: da.Array
        Sign multiplier that ensures deterministic output.
    """
    max_vals = np.max(data, axis=axis)
    min_vals = np.min(data, axis=axis)
    sign_multiplier = np.where(np.abs(max_vals) >= np.abs(min_vals), 1, -1)
    return sign_multiplier


class _SVD:
    """Decomposes a data object using Singular Value Decomposition (SVD).

    The data object will be decomposed like X = U * S * V.T, where U and V are
    orthogonal matrices and S is a diagonal matrix with the singular values on
    the diagonal.

    Parameters
    ----------
    n_modes : int | float | str
        Number of components to be computed. If float, it represents the fraction of variance to keep. If "all", all components are kept.
    init_rank_reduction : float, default=0.0
        Used only when `n_modes` is given as a float. Specifiy the initial target rank to be computed by randomized SVD before truncating the solution to the desired fraction of explained variance. Must be in the half open interval (0, 1]. Lower values will speed up the computation. If the rank is too low and the fraction of explained variance is not reached, a warning will be raised.
    flip_signs : bool, default=True
        Whether to flip the sign of the components to ensure deterministic output.
    solver : {'auto', 'full', 'randomized'}, default='auto'
        The solver is selected by a default policy based on size of `X` and `n_modes`:
        if the input data is larger than 500x500 and the number of modes to extract is lower
        than 80% of the smallest dimension of the data, then the more efficient
        `randomized` method is enabled. Otherwise the exact full SVD is computed
        and optionally truncated afterwards.
    random_state : np.random.Generator | int | None, default=None
        Seed for the random number generator.
    solver_kwargs : dict, default={}
        Additional keyword arguments passed to the SVD solver.
    """

    def __init__(
        self,
        n_modes: int | float | str,
        init_rank_reduction: float = 0.3,
        flip_signs: bool = True,
        solver: str = "auto",
        random_state: np.random.Generator | int | None = None,
        solver_kwargs: dict = {},
        is_complex: bool | str = "auto",
    ):
        sanity_check_n_modes(n_modes)
        self.is_based_on_variance = True if isinstance(n_modes, float) else False

        if self.is_based_on_variance:
            if not (0 < init_rank_reduction <= 1.0):
                raise ValueError(
                    "init_rank_reduction must be in the half open interval (0, 1]."
                )

        self.n_modes = n_modes
        self.n_modes_precompute = n_modes
        self.init_rank_reduction = init_rank_reduction
        self.flip_signs = flip_signs
        self.solver = solver
        self.random_state = random_state
        self.solver_kwargs = solver_kwargs
        self.is_complex = is_complex

    def _get_n_modes_precompute(self, rank: int) -> int:
        if self.is_based_on_variance:
            n_modes_precompute = int(rank * self.init_rank_reduction)
            if n_modes_precompute < 1:
                warnings.warn(
                    f"`init_rank_reduction={self.init_rank_reduction}` is too low resulting in zero components. One component will be computed instead."
                )
                n_modes_precompute = 1
        elif self.n_modes_precompute == "all":
            n_modes_precompute = rank

        elif isinstance(self.n_modes_precompute, int):
            if self.n_modes_precompute > rank:
                raise ValueError(
                    f"n_modes must be less than or equal to the rank of the dataset (rank = {rank})."
                )
            n_modes_precompute = self.n_modes_precompute
        return n_modes_precompute

    def fit_transform(self, X):
        """Decomposes the data object.

        Parameters
        ----------
        X : DataArray
            A 2-dimensional data object to be decomposed.
        """
        rank = min(X.shape)
        self.n_modes_precompute = self._get_n_modes_precompute(rank)

        # Check if data is small enough to use exact SVD
        # If not, use randomized SVD
        # If data is complex, use scipy sparse SVD
        # If data is dask, use dask SVD
        # Conditions for using exact SVD follow scitkit-learn's PCA implementation
        # Source: https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html

        use_dask = True if isinstance(X, DaskArray) else False

        match self.is_complex:
            case bool():
                use_complex = self.is_complex
            case "auto":
                use_complex = True if np.iscomplexobj(X) else False
            case _:
                raise ValueError(
                    f"Unrecognized value for is_complex '{self.is_complex}'. "
                    "Valid options are True, False, and 'auto'."
                )

        is_small_data = max(X.shape) < 500

        match self.solver:
            # TODO(nicrie): implement more performant case for tall and skinny problems which are best handled by precomputing the covariance matrix.
            # if X.shape[1] <= 1_000 and X.shape[0] >= 10 * X.shape[1]: -> covariance_eigh" (see sklearn PCA implementation: https://github.com/scikit-learn/scikit-learn/blob/e87b32a81c70abed8f2e97483758eb64df8255e9/sklearn/decomposition/_pca.py#L526)
            case "auto":
                has_many_modes = self.n_modes_precompute > int(0.8 * rank)
                use_exact = (
                    True if is_small_data and has_many_modes and not use_dask else False
                )
            case "full":
                use_exact = True
            case "randomized":
                use_exact = False
            case _:
                raise ValueError(
                    f"Unrecognized solver '{self.solver}'. "
                    "Valid options are 'auto', 'full', and 'randomized'."
                )

        # Use exact SVD for small data sets
        if use_exact:
            U, s, VT = self._svd(X, np.linalg.svd, self.solver_kwargs)
            U = U[:, : self.n_modes_precompute]
            s = s[: self.n_modes_precompute]
            VT = VT[: self.n_modes_precompute, :]

        # Use randomized SVD for large, real-valued data sets
        elif (not use_complex) and (not use_dask):
            solver_kwargs = self.solver_kwargs | {
                "n_components": self.n_modes_precompute,
                "random_state": self.random_state,
            }
            U, s, VT = self._svd(X, randomized_svd, solver_kwargs)

        # Use scipy sparse SVD for large, complex-valued data sets
        elif use_complex and (not use_dask):
            # Scipy sparse version
            solver_kwargs = self.solver_kwargs | {
                "k": self.n_modes_precompute,
                "solver": "lobpcg",
                "random_state": self.random_state,
            }
            U, s, VT = self._svd(X, complex_svd, solver_kwargs)
            idx_sort = np.argsort(s)[::-1]
            U = U[:, idx_sort]
            s = s[idx_sort]
            VT = VT[idx_sort, :]

        # Use dask SVD for large, real-valued, delayed data sets
        elif (not use_complex) and use_dask:
            solver_kwargs = self.solver_kwargs | {
                "k": self.n_modes_precompute,
                "seed": self.random_state,
            }
            solver_kwargs.setdefault("compute", False)
            solver_kwargs.setdefault("n_power_iter", 4)
            U, s, VT = self._svd(X, dask_svd, solver_kwargs)
        else:
            err_msg = (
                "Complex data together with dask is currently not implemented. See dask issue 7639 "
                "https://github.com/dask/dask/issues/7639"
            )
            raise NotImplementedError(err_msg)

        U, s, VT = wait_on(U, s, VT)

        V = VT.conj().T

        # Flip signs of components to ensure deterministic output
        if self.flip_signs:
            sign_multiplier = get_deterministic_sign_multiplier(V, axis=0)
            V *= sign_multiplier
            U *= sign_multiplier

        # Truncate the decomposition to the desired number of modes
        if self.is_based_on_variance:
            # Truncating based on variance requires computation of dask array
            # which we prefer to avoid
            if use_dask:
                err_msg = "Estimating the number of modes to keep based on variance is not supported with dask arrays. Please explicitly specifiy the number of modes to keep by using an integer for the number of modes."
                raise ValueError(err_msg)

            # Compute the fraction of explained variance per mode
            N = X.shape[0] - 1
            total_variance = X.var(axis=0, ddof=1).sum()
            explained_variance = s**2 / N / total_variance
            cum_expvar = explained_variance.cumsum()
            total_explained_variance = cum_expvar[-1].item()

            n_modes_required = (
                self.n_modes_precompute - (cum_expvar >= self.n_modes).sum() + 1
            )
            if n_modes_required > self.n_modes_precompute:
                warnings.warn(
                    f"Dataset has {self.n_modes_precompute} components, explaining {total_explained_variance:.2%} of the variance. However, {self.n_modes:.2%} explained variance was requested. Please consider increasing `init_rank_reduction`."
                )
                n_modes_required = self.n_modes_precompute

            # Truncate solution to the desired fraction of explained variance
            U = U[:, :n_modes_required]
            s = s[:n_modes_required]
            V = V[:, :n_modes_required]

        return U, s, V

    def _svd(self, X, func, kwargs):
        """Performs SVD on the data

        Parameters
        ----------
        X : DataArray
            A 2-dimensional data object to be decomposed.
        func : Callable
            Method to perform SVD.
        kwargs : dict
            Additional keyword arguments passed to the SVD solver.

        Returns
        -------
        U : DataArray
            Left singular vectors.
        s : DataArray
            Singular values.
        VT : DataArray
            Right singular vectors.
        """
        try:
            U, s, VT = func(X, **kwargs)
            return U, s, VT
        except np.linalg.LinAlgError:
            raise np.linalg.LinAlgError(
                "SVD failed. This may be due to isolated NaN values in the data. Please consider the following steps:\n"
                "1. Check for and remove any isolated NaNs in your dataset.\n"
                "2. If the error persists, please raise an issue at https://github.com/xarray-contrib/xeofs/issues."
            )



import warnings
from typing import Sequence

import numpy as np
import xarray as xr
from typing_extensions import Self

from ..linalg._numpy import _fractional_matrix_power
from ..linalg.decomposer import Decomposer
from ..utils.data_types import DataArray, DataObject
from ..utils.hilbert_transform import hilbert_transform
from ..utils.xarray_utils import argsort_dask
from .base_model_cross_set import BaseModelCrossSet


class CPCCA(BaseModelCrossSet):
    """Continuum Power CCA (CPCCA).

    CPCCA extends continuum power regression to isolate pairs of coupled
    patterns, maximizing the squared covariance between partially whitened
    variables [1]_ [2]_.

    This method solves the following optimization problem:

        :math:`\\max_{q_x, q_y} \\left( q_x^T X^T Y q_y \\right)`

    subject to the constraints:

        :math:`q_x^T (X^TX)^{1-\\alpha_x} q_x = 1, \\quad q_y^T
        (Y^TY)^{1-\\alpha_y} q_y = 1`

    where :math:`\\alpha_x` and :math:`\\alpha_y` control the degree of
    whitening applied to the data.

    Parameters
    ----------
    n_modes : int, default=2
        Number of modes to calculate.
    alpha : Sequence[float] | float, default=0.2
        Degree of whitening applied to the data. If float, the same value is
        applied to both data sets.
    standardize : Squence[bool] | bool, default=False
        Whether to standardize the input data. Generally not recommended as
        standardization can be managed by the degree of whitening.
    use_coslat : Sequence[bool] | bool, default=False
        For data on a longitude-latitude grid, whether to correct for varying
        grid cell areas towards the poles by scaling each grid point with the
        square root of the cosine of its latitude.
    use_pca : Sequence[bool] | bool, default=False
        Whether to preprocess each field individually by reducing dimensionality
        through PCA. The cross-covariance matrix is then computed in the reduced
        principal component space.
    n_pca_modes : Sequence[int | float | str] | int | float | str, default=0.999
        Number of modes to retain during PCA preprocessing step. If int,
        specifies the exact number of modes; if float, specifies the fraction of
        variance to retain; if "all", all modes are retained.
    pca_init_rank_reduction : Sequence[float] | float, default=0.3
        Relevant when `use_pca=True` and `n_pca_modes` is a float. Specifies the
        initial fraction of rank reduction for faster PCA computation via
        randomized SVD.
    check_nans : Sequence[bool] | bool, default=True
        Whether to check for NaNs in the input data. Set to False for lazy model
        evaluation.
    compute : bool, default=True
        Whether to compute the model elements eagerly. If True, the following
        are computed sequentially: preprocessor scaler, optional NaN checks, SVD
        decomposition, scores, and components.
    random_state : numpy.random.Generator | int | None, default=None
        Seed for the random number generator.
    sample_name : str, default="sample"
        Name for the new sample dimension.
    feature_name : Sequence[str] | str, default="feature"
        Name for the new feature dimension.
    solver : {"auto", "full", "randomized"}
        Solver to use for the SVD computation.
    solver_kwargs : dict, default={}
        Additional keyword arguments passed to the SVD solver function.

    Notes
    -----
    Canonical Correlation Analysis (CCA), Maximum Covariance Analysis (MCA) and
    Redundany Analysis (RDA) are all special cases of CPCCA depending on the
    choice of the parameter :math:`\\alpha`.

    References
    ----------
    .. [1] Swenson, E. Continuum Power CCA: A Unified Approach for Isolating
        Coupled Modes. Journal of Climate 28, 1016–1030 (2015).
    .. [2] Wilks, D. S. Statistical Methods in the Atmospheric Sciences.
        (Academic Press, 2019).
        doi:https://doi.org/10.1016/B978-0-12-815823-4.00011-0.

    Examples
    --------
    Perform regular CCA on two data sets:

    >>> model = CPCCA(n_modes=5, alpha=0.0)
    >>> model.fit(X, Y)

    Perform regularized CCA on two data sets:

    >>> model = CPCCA(n_modes=5, alpha=0.2)
    >>> model.fit(X, Y)

    Perform Maximum Covariance Analysis:

    >>> model = CPCCA(n_modes=5, alpha=1.0)
    >>> model.fit(X, Y)

    Perform Redundancy Analysis:

    >>> model = CPCCA(n_modes=5, alpha=[0, 1])
    >>> model.fit(X, Y)

    Make predictions for `Y` given `X`:

    >>> scores_y_pred = model.predict(X)  # prediction in "PC" space
    >>> Y_pred = model.inverse_transform(Y=scores_y_pred)  # prediction in physical space


    """

    extra_modules = ["statsmodels"]

    def __init__(
        self,
        n_modes: int = 2,
        alpha: Sequence[float] | float = 0.2,
        standardize: Sequence[bool] | bool = False,
        use_coslat: Sequence[bool] | bool = False,
        use_pca: Sequence[bool] | bool = True,
        n_pca_modes: Sequence[float | int | str] | float | int | str = 0.999,
        pca_init_rank_reduction: Sequence[float] | float = 0.3,
        check_nans: Sequence[bool] | bool = True,
        compute: bool = True,
        sample_name: str = "sample",
        feature_name: Sequence[str] | str = "feature",
        solver: str = "auto",
        random_state: np.random.Generator | int | None = None,
        solver_kwargs: dict = {},
        **kwargs,
    ):
        super().__init__(
            n_modes=n_modes,
            center=True,
            standardize=standardize,
            use_coslat=use_coslat,
            check_nans=check_nans,
            use_pca=use_pca,
            n_pca_modes=n_pca_modes,
            pca_init_rank_reduction=pca_init_rank_reduction,
            alpha=alpha,
            compute=compute,
            sample_name=sample_name,
            feature_name=feature_name,
            solver=solver,
            random_state=random_state,
            solver_kwargs=solver_kwargs,
        )
        self.attrs.update({"model": "Continuum Power CCA"})
        # Remove center from the inherited serialization params because it is hard-coded for CPCCA
        self._params.pop("center")

        params = self.get_params()
        self.sample_name: str = params["sample_name"]
        self.feature_name: tuple[str, str] = params["feature_name"]

    def _fit_algorithm(
        self,
        X: DataArray,
        Y: DataArray,
    ) -> Self:
        feature_name = self.feature_name

        # Compute the totalsquared covariance from the unwhitened data
        C_whitened = self._compute_cross_matrix(
            X,
            Y,
            sample_dim=self.sample_name,
            feature_dim_x=feature_name[0],
            feature_dim_y=feature_name[1],
            method="covariance",
            diagonal=False,
        )

        # Initialize the SVD decomposer
        decomposer = Decomposer(**self._decomposer_kwargs)
        dims = (feature_name[0], feature_name[1])
        decomposer.fit(C_whitened, dims=dims)

        # Store the results
        singular_values = decomposer.s_
        Q1 = decomposer.U_
        Q2 = decomposer.V_

        # Compute total squared variance
        total_squared_covariance = self._compute_total_squared_covariance(C_whitened)

        # Index of the sorted covariance explained
        idx_sorted_modes = argsort_dask(singular_values, "mode")[::-1]  # type: ignore
        idx_sorted_modes.coords.update(singular_values.coords)

        # Project the data onto the singular vectors
        scores1 = xr.dot(X, Q1, dims=feature_name[0])
        scores2 = xr.dot(Y, Q2, dims=feature_name[1])

        norm1 = np.sqrt(xr.dot(scores1.conj(), scores1, dims=self.sample_name)).real
        norm2 = np.sqrt(xr.dot(scores2.conj(), scores2, dims=self.sample_name)).real

        self.data.add(name="input_data1", data=X, allow_compute=False)
        self.data.add(name="input_data2", data=Y, allow_compute=False)
        self.data.add(name="components1", data=Q1)
        self.data.add(name="components2", data=Q2)
        self.data.add(name="scores1", data=scores1)
        self.data.add(name="scores2", data=scores2)
        self.data.add(name="singular_values", data=singular_values)
        self.data.add(name="squared_covariance", data=singular_values**2)
        self.data.add(name="total_squared_covariance", data=total_squared_covariance)
        self.data.add(name="idx_modes_sorted", data=idx_sorted_modes)
        self.data.add(name="norm1", data=norm1)
        self.data.add(name="norm2", data=norm2)

        # # Assign analysis-relevant meta data
        self.data.set_attrs(self.attrs)
        return self

    def _transform_algorithm(
        self,
        X: DataArray | None = None,
        Y: DataArray | None = None,
        normalized=False,
    ) -> dict[str, DataArray]:
        results = {}
        if X is not None:
            # Project data onto singular vectors
            comps1 = self.data["components1"]
            norm1 = self.data["norm1"]
            scores1 = xr.dot(X, comps1)
            if normalized:
                scores1 = scores1 / norm1
            results["X"] = scores1

        if Y is not None:
            # Project data onto singular vectors
            comps2 = self.data["components2"]
            norm2 = self.data["norm2"]
            scores2 = xr.dot(Y, comps2)
            if normalized:
                scores2 = scores2 / norm2
            results["Y"] = scores2

        return results

    def _inverse_transform_algorithm(
        self, X: DataArray | None = None, Y: DataArray | None = None
    ) -> dict[str, DataArray]:
        x_is_given = X is not None
        y_is_given = Y is not None

        if (not x_is_given) and (not y_is_given):
            raise ValueError("Either X or Y must be given.")

        results = {}

        if x_is_given:
            # Singular vectors
            comps1 = self.data["components1"].sel(mode=X.mode)
            # Reconstruct the data
            results["X"] = xr.dot(X, comps1.conj(), dims="mode")

        if y_is_given:
            # Singular vectors
            comps2 = self.data["components2"].sel(mode=Y.mode)
            # Reconstruct the data
            results["Y"] = xr.dot(Y, comps2.conj(), dims="mode")

        return results

    def _predict_algorithm(self, X: DataArray) -> DataArray:
        sample_name_fit_x = "sample_fit_dim_x"
        sample_name_fit_y = "sample_fit_dim_y"
        Qx = self.data["components1"]
        Rx = self.data["scores1"].rename({self.sample_name: sample_name_fit_x})
        Ry = self.data["scores2"].rename({self.sample_name: sample_name_fit_y})

        def _predict_numpy(X, Qx, Rx, Ry):
            G = Rx.conj().T @ Ry / np.linalg.norm(Rx, axis=0) ** 2
            return X @ Qx @ G

        Ry_pred = xr.apply_ufunc(
            _predict_numpy,
            X,
            Qx,
            Rx,
            Ry,
            input_core_dims=[
                [self.sample_name, self.feature_name[0]],
                [self.feature_name[0], "mode"],
                [sample_name_fit_x, "mode"],
                [sample_name_fit_y, "mode"],
            ],
            output_core_dims=[[self.sample_name, "mode"]],
            dask="allowed",
        )
        Ry_pred.name = "pseudo_scores_Y"
        return Ry_pred

    def _get_components(self, normalized=True):
        comps1 = self.data["components1"]
        comps2 = self.data["components2"]

        if not normalized:
            comps1 = comps1 * self.data["norm1"]
            comps2 = comps2 * self.data["norm2"]

        return comps1, comps2

    def _get_scores(self, normalized=False):
        norm1 = self.data["norm1"]
        norm2 = self.data["norm2"]

        scores1 = self.data["scores1"]
        scores2 = self.data["scores2"]

        if normalized:
            scores1 = scores1 / norm1
            scores2 = scores2 / norm2

        return scores1, scores2

    def cross_correlation_coefficients(self):
        """Get the cross-correlation coefficients.

        The cross-correlation coefficients between the scores of ``X`` and ``Y``
        are computed as:

        .. math::
            c_{xy, i} = \\text{corr} \\left(\\mathbf{r}_{x, i}, \\mathbf{r}_{y, i} \\right)

        where :math:`\\mathbf{r}_{x, i}` and :math:`\\mathbf{r}_{y, i}` are the
        `i`th scores of ``X`` and ``Y``,

        Notes
        -----
        When :math:`\\alpha=0`, the cross-correlation coefficients are
        equivalent to the canonical correlation coefficients.

        """

        Rx = self.data["scores1"]
        Ry = self.data["scores2"]

        cross_corr = self._compute_cross_matrix(
            Rx,
            Ry,
            sample_dim=self.sample_name,
            feature_dim_x="mode",
            feature_dim_y="mode",
            method="correlation",
            diagonal=True,
        )
        cross_corr = cross_corr.real
        cross_corr.name = "cross_correlation_coefficients"
        return cross_corr

    def correlation_coefficients_X(self):
        """Get the correlation coefficients for the scores of :math:`X`.

        The correlation coefficients of the scores of :math:`X` are given by:

        .. math::
            c_{x, ij} = \\text{corr} \\left(\\mathbf{r}_{x, i}, \\mathbf{r}_{x, j} \\right)

        where :math:`\\mathbf{r}_{x, i}` and :math:`\\mathbf{r}_{x, j}` are the
        `i`th and `j`th scores of :math:`X`.

        """
        Rx = self.data["scores1"]

        corr = self._compute_cross_matrix(
            Rx,
            Rx,
            sample_dim=self.sample_name,
            feature_dim_x="mode",
            feature_dim_y="mode",
            method="correlation",
            diagonal=False,
        )
        corr.name = "correlation_coefficients_X"
        return corr

    def correlation_coefficients_Y(self):
        """Get the correlation coefficients for the scores of :math:`Y`.

        The correlation coefficients of the scores of :math:`Y` are given by:

        .. math::
            c_{y, ij} = \\text{corr} \\left(\\mathbf{r}_{y, i}, \\mathbf{r}_{y, j} \\right)

        where :math:`\\mathbf{r}_{y, i}` and :math:`\\mathbf{r}_{y, j}` are the
        `i`th and `j`th scores of :math:`Y`.

        """
        Ry = self.data["scores2"]

        corr = self._compute_cross_matrix(
            Ry,
            Ry,
            sample_dim=self.sample_name,
            feature_dim_x="mode",
            feature_dim_y="mode",
            method="correlation",
            diagonal=False,
        )
        corr.name = "correlation_coefficients_Y"
        return corr

    def squared_covariance_fraction(self):
        """Get the squared covariance fraction (SCF).

        The SCF is computed as a weighted mean-square error (see equation (15)
        in Swenson (2015)) :

        .. math::
            SCF_{i} = 1 - \\frac{\\|\\mathbf{d}_{X,i}^T \\mathbf{d}_{Y,i}\\|_F^2}{\\|X^TY\\|_F^2}

        where :math:`\\mathbf{d}_{X,i}` and :math:`\\mathbf{d}_{Y,i}` are the
        residuals of the input data :math:`X` and :math:`Y` after reconstruction
        by the `ith` scores of :math:`X` and :math:`Y`, respectively.

        References
        ----------
        Swenson, E. Continuum Power CCA: A Unified Approach for Isolating
            Coupled Modes. Journal of Climate 28, 1016–1030 (2015).

        """

        def _compute_residual_variance_numpy(X, Y, Xrec, Yrec):
            dX = X - Xrec
            dY = Y - Yrec

            return np.linalg.norm(dX.conj().T @ dY / (dX.shape[0] - 1)) ** 2

        total_squared_covariance = self.data["total_squared_covariance"]
        sample_name_x = "sample_dim_x"
        sample_name_y = "sample_dim_y"

        # Get the singular vectors
        Q1 = self.data["components1"]
        Q2 = self.data["components2"]

        # Get input data
        X1 = self.data["input_data1"]
        X2 = self.data["input_data2"]

        # Unwhiten the data
        X1 = self.whitener1.inverse_transform_data(X1)
        X2 = self.whitener2.inverse_transform_data(X2)

        # Rename the sample dimension to avoid conflicts for
        # different coordinates with same length
        X1 = X1.rename({self.sample_name: sample_name_x})
        X2 = X2.rename({self.sample_name: sample_name_y})

        # Get the component scores
        scores1 = self.data["scores1"]
        scores2 = self.data["scores2"]

        # Compute the residual variance for each mode
        squared_covariance_fraction: list[DataArray] = []
        for mode in scores1.mode.values:
            # Reconstruct the data
            X1r = xr.dot(
                scores1.sel(mode=[mode]), Q1.sel(mode=[mode]).conj().T, dims="mode"
            )
            X2r = xr.dot(
                scores2.sel(mode=[mode]), Q2.sel(mode=[mode]).conj().T, dims="mode"
            )

            # Unwhitend the reconstructed data
            X1r = self.whitener1.inverse_transform_data(X1r)
            X2r = self.whitener2.inverse_transform_data(X2r)

            # Compute fraction variance explained
            X1r = X1r.rename({self.sample_name: sample_name_x})
            X2r = X2r.rename({self.sample_name: sample_name_y})
            res_var: DataArray = xr.apply_ufunc(
                _compute_residual_variance_numpy,
                X1,
                X2,
                X1r,
                X2r,
                input_core_dims=[
                    [sample_name_x, self.feature_name[0]],
                    [sample_name_y, self.feature_name[1]],
                    [sample_name_x, self.feature_name[0]],
                    [sample_name_y, self.feature_name[1]],
                ],
                output_core_dims=[[]],
                dask="allowed",
            )
            res_var = res_var.expand_dims({"mode": [mode]})
            squared_covariance_fraction.append(1 - res_var / total_squared_covariance)

        scf = xr.concat(squared_covariance_fraction, dim="mode")
        scf.name = "squared_covariance_fraction"

        # In theory, the residual can be larger than the total squared covariance
        # if a mode is not well defined. In this case, the SCF would be negative.
        # We set these values to zero.
        scf = xr.where(scf < 0, 0, scf)
        return scf

    def fraction_variance_X_explained_by_X(self):
        """Get the fraction of variance explained (FVE X).

        The FVE X is the fraction of variance in :math:`X` explained by the
        scores of :math:`X`. It is computed as a weighted mean-square error (see
        equation (15) in Swenson (2015)) :

        .. math::
            FVE_{X|X,i} = 1 - \\frac{\\|\\mathbf{d}_{X,i}\\|_F^2}{\\|X\\|_F^2}

        where :math:`\\mathbf{d}_{X,i}` are the residuals of the input data
        :math:`X` after reconstruction by the `ith` scores of :math:`X`.

        References
        ----------
        Swenson, E. Continuum Power CCA: A Unified Approach for Isolating
            Coupled Modes. Journal of Climate 28, 1016–1030 (2015).

        """
        # Get the singular vectors
        Qx = self.data["components1"]

        # Get input data
        X = self.data["input_data1"]

        # Unwhiten the data
        X = self.whitener1.inverse_transform_data(X)

        # Compute the total variance
        total_variance: DataArray = self._compute_total_variance(X, self.sample_name)

        # Get the component scores
        Rx = self.data["scores1"]

        # Compute the residual variance for each mode
        fraction_variance_explained: list[DataArray] = []
        for mode in Rx.mode.values:
            # Reconstruct the data
            Xr = xr.dot(Rx.sel(mode=[mode]), Qx.sel(mode=[mode]).conj().T, dims="mode")

            # Unwhitend the reconstructed data
            Xr = self.whitener1.inverse_transform_data(Xr)

            # Compute fraction variance explained
            residual_variance = self._compute_total_variance(X - Xr, self.sample_name)
            residual_variance = residual_variance.expand_dims({"mode": [mode]})
            fraction_variance_explained.append(1 - residual_variance / total_variance)

        fve_xx = xr.concat(fraction_variance_explained, dim="mode")
        fve_xx.name = "fraction_variance_X_explained_by_X"
        return fve_xx

    def fraction_variance_Y_explained_by_Y(self):
        """Get the fraction of variance explained (FVE Y).

        The FVE Y is the fraction of variance in :math:`Y` explained by the
        scores of :math:`Y`. It is computed as a weighted mean-square error (see
        equation (15) in Swenson (2015)) :

        .. math::
            FVE_{Y|Y,i} = 1 - \\frac{\\|\\mathbf{d}_{Y,i}\\|_F^2}{\\|Y\\|_F^2}

        where :math:`\\mathbf{d}_{Y,i}` are the residuals of the input data
        :math:`Y` after reconstruction by the `ith` scores of :math:`Y`.

        References
        ----------
        Swenson, E. Continuum Power CCA: A Unified Approach for Isolating
            Coupled Modes. Journal of Climate 28, 1016–1030 (2015).

        """
        # Get the singular vectors
        Qy = self.data["components2"]

        # Get input data
        Y = self.data["input_data2"]

        # Unwhiten the data
        Y = self.whitener2.inverse_transform_data(Y)

        # Compute the total variance
        total_variance: DataArray = self._compute_total_variance(Y, self.sample_name)

        # Get the component scores
        Ry = self.data["scores2"]

        # Compute the residual variance for each mode
        fraction_variance_explained: list[DataArray] = []
        for mode in Ry.mode.values:
            # Reconstruct the data
            Yr = xr.dot(Ry.sel(mode=[mode]), Qy.sel(mode=[mode]).conj().T, dims="mode")

            # Unwhitend the reconstructed data
            Yr = self.whitener2.inverse_transform_data(Yr)

            # Compute fraction variance explained
            residual_variance = self._compute_total_variance(Y - Yr, self.sample_name)
            residual_variance = residual_variance.expand_dims({"mode": [mode]})
            fraction_variance_explained.append(1 - residual_variance / total_variance)

        fve_yy = xr.concat(fraction_variance_explained, dim="mode")
        fve_yy.name = "fraction_variance_Y_explained_by_Y"
        return fve_yy

    def fraction_variance_Y_explained_by_X(self) -> DataArray:
        """Get the fraction of variance explained (FVE YX).

        The FVE YX is the fraction of variance in :math:`Y` explained by the
        scores of :math:`X`. It is computed as a weighted mean-square error (see
        equation (15) in Swenson (2015)) :

        .. math::
            FVE_{Y|X,i} = 1 - \\frac{\\|(X^TX)^{-1/2} \\mathbf{d}_{X,i}^T \\mathbf{d}_{Y,i}\\|_F^2}{\\|(X^TX)^{-1/2} X^TY\\|_F^2}

        where :math:`\\mathbf{d}_{X,i}` and :math:`\\mathbf{d}_{Y,i}` are the
        residuals of the input data :math:`X` and :math:`Y` after reconstruction
        by the `ith` scores of :math:`X` and :math:`Y`, respectively.

        References
        ----------
        Swenson, E. Continuum Power CCA: A Unified Approach for Isolating
        Coupled Modes. Journal of Climate 28, 1016–1030 (2015).

        """

        def _compute_total_variance_numpy(X, Y):
            Cx = X.conj().T @ X / (X.shape[0] - 1)
            Tinv = _fractional_matrix_power(Cx, -0.5)
            return np.linalg.norm(Tinv @ X.conj().T @ Y / (X.shape[0] - 1)) ** 2

        def _compute_residual_variance_numpy(X, Y, Xrec, Yrec):
            dX = X - Xrec
            dY = Y - Yrec

            Cx = X.conj().T @ X / (X.shape[0] - 1)
            Tinv = _fractional_matrix_power(Cx, -0.5)
            return np.linalg.norm(Tinv @ dX.conj().T @ dY / (dX.shape[0] - 1)) ** 2

        sample_name_x = "sample_dim_x"
        sample_name_y = "sample_dim_y"

        # Get the singular vectors
        Q1 = self.data["components1"]
        Q2 = self.data["components2"]

        # Get input data
        X1 = self.data["input_data1"]
        X2 = self.data["input_data2"]

        # Unwhiten the data
        X1 = self.whitener1.inverse_transform_data(X1)
        X2 = self.whitener2.inverse_transform_data(X2)

        # Compute the total variance
        X1 = X1.rename({self.sample_name: sample_name_x})
        X2 = X2.rename({self.sample_name: sample_name_y})
        total_variance: DataArray = xr.apply_ufunc(
            _compute_total_variance_numpy,
            X1,
            X2,
            input_core_dims=[
                [sample_name_x, self.feature_name[0]],
                [sample_name_y, self.feature_name[1]],
            ],
            output_core_dims=[[]],
            dask="allowed",
        )

        # Get the component scores
        scores1 = self.data["scores1"]
        scores2 = self.data["scores2"]

        # Compute the residual variance for each mode
        fraction_variance_explained: list[DataArray] = []
        for mode in scores1.mode.values:
            # Reconstruct the data
            X1r = xr.dot(
                scores1.sel(mode=[mode]), Q1.sel(mode=[mode]).conj().T, dims="mode"
            )
            X2r = xr.dot(
                scores2.sel(mode=[mode]), Q2.sel(mode=[mode]).conj().T, dims="mode"
            )

            # Unwhitend the reconstructed data
            X1r = self.whitener1.inverse_transform_data(X1r)
            X2r = self.whitener2.inverse_transform_data(X2r)

            # Compute fraction variance explained
            X1r = X1r.rename({self.sample_name: sample_name_x})
            X2r = X2r.rename({self.sample_name: sample_name_y})
            res_var: DataArray = xr.apply_ufunc(
                _compute_residual_variance_numpy,
                X1,
                X2,
                X1r,
                X2r,
                input_core_dims=[
                    [sample_name_x, self.feature_name[0]],
                    [sample_name_y, self.feature_name[1]],
                    [sample_name_x, self.feature_name[0]],
                    [sample_name_y, self.feature_name[1]],
                ],
                output_core_dims=[[]],
                dask="allowed",
            )
            res_var = res_var.expand_dims({"mode": [mode]})
            fraction_variance_explained.append(1 - res_var / total_variance)

        fve_yx = xr.concat(fraction_variance_explained, dim="mode")
        fve_yx.name = "fraction_variance_Y_explained_by_X"
        return fve_yx

    def homogeneous_patterns(self, correction=None, alpha=0.05):
        """Get the homogeneous correlation patterns.

        The homogeneous correlation patterns are the correlation coefficients
        between the input data and the scores. They are defined as:

        .. math::
            H_{X, i} = \\text{corr} \\left(X, \\mathbf{r}_{x,i} \\right)

        .. math::
            H_{Y, i} = \\text{corr} \\left(Y, \\mathbf{r}_{y,i} \\right)

        where :math:`X` and :math:`Y` are the input data, and
        :math:`\\mathbf{r}_{x,i}` and :math:`\\mathbf{r}_{y,i}` are the `i`th
        scores of :math:`X` and :math:`Y`, respectively.


        Parameters
        ----------
        correction: str, default=None
            Method to apply a multiple testing correction. If None, no
            correction is applied.  Available methods are: - bonferroni :
            one-step correction - sidak : one-step correction - holm-sidak :
            step down method using Sidak adjustments - holm : step-down method
            using Bonferroni adjustments - simes-hochberg : step-up method
            (independent) - hommel : closed method based on Simes tests
            (non-negative) - fdr_bh : Benjamini/Hochberg (non-negative)
            (default) - fdr_by : Benjamini/Yekutieli (negative) - fdr_tsbh : two
            stage fdr correction (non-negative) - fdr_tsbky : two stage fdr
            correction (non-negative)
        alpha: float, default=0.05
            The desired family-wise error rate. Not used if `correction` is
            None.

        Returns
        -------
        tuple[DataObject, DataObject]
            Homogenous correlation patterns of `X` and `Y`.
        tuple[DataObject, DataObject]
            p-values of the homogenous correlation patterns of `X` and `Y`.

        """
        from ..utils.optional.statistics import pearson_correlation

        input_data1 = self.data["input_data1"]
        input_data2 = self.data["input_data2"]

        input_data1 = self.whitener1.inverse_transform_data(input_data1)
        input_data2 = self.whitener2.inverse_transform_data(input_data2)

        input_data1 = self.pca1.inverse_transform_data(input_data1)
        input_data2 = self.pca2.inverse_transform_data(input_data2)

        scores1 = self.data["scores1"]
        scores2 = self.data["scores2"]

        hom_pat1, pvals1 = pearson_correlation(
            input_data1,
            scores1,
            correction=correction,
            alpha=alpha,
            sample_name=self.sample_name,
            feature_name=self.feature_name[0],
        )
        hom_pat2, pvals2 = pearson_correlation(
            input_data2,
            scores2,
            correction=correction,
            alpha=alpha,
            sample_name=self.sample_name,
            feature_name=self.feature_name[1],
        )

        hom_pat1.name = "left_homogeneous_patterns"
        hom_pat2.name = "right_homogeneous_patterns"

        pvals1.name = "pvalues_of_left_homogeneous_patterns"
        pvals2.name = "pvalues_of_right_homogeneous_patterns"

        hom_pat1 = self.preprocessor1.inverse_transform_components(hom_pat1)
        hom_pat2 = self.preprocessor2.inverse_transform_components(hom_pat2)

        pvals1 = self.preprocessor1.inverse_transform_components(pvals1)
        pvals2 = self.preprocessor2.inverse_transform_components(pvals2)

        return (hom_pat1, hom_pat2), (pvals1, pvals2)

    def heterogeneous_patterns(self, correction=None, alpha=0.05):
        """Get the heterogeneous correlation patterns.

        The heterogeneous patterns are the correlation coefficients between the
        input data and the scores of the other field:

        .. math::
            G_{X, i} = \\text{corr} \\left(X, \\mathbf{r}_{y,i} \\right)

        .. math::
            G_{Y, i} = \\text{corr} \\left(Y, \\mathbf{r}_{x,i} \\right)

        where :math:`X` and :math:`Y` are the input data, and
        :math:`\\mathbf{r}_{x,i}` and :math:`\\mathbf{r}_{y,i}` are the `i`th
        scores of :math:`X` and :math:`Y`, respectively.

        Parameters
        ----------
        correction: str, default=None
            Method to apply a multiple testing correction. If None, no
            correction is applied.  Available methods are: - bonferroni :
            one-step correction - sidak : one-step correction - holm-sidak :
            step down method using Sidak adjustments - holm : step-down method
            using Bonferroni adjustments - simes-hochberg : step-up method
            (independent) - hommel : closed method based on Simes tests
            (non-negative) - fdr_bh : Benjamini/Hochberg (non-negative)
            (default) - fdr_by : Benjamini/Yekutieli (negative) - fdr_tsbh : two
            stage fdr correction (non-negative) - fdr_tsbky : two stage fdr
            correction (non-negative)
        alpha: float, default=0.05
            The desired family-wise error rate. Not used if `correction` is
            None.

        Returns
        -------
        tuple[DataObject, DataObject]
            Heterogenous correlation patterns of `X` and `Y`.
        tuple[DataObject, DataObject]
            p-values of the heterogenous correlation patterns of `X` and `Y`.

        """
        from ..utils.optional.statistics import pearson_correlation

        input_data1 = self.data["input_data1"]
        input_data2 = self.data["input_data2"]

        input_data1 = self.whitener1.inverse_transform_data(input_data1)
        input_data2 = self.whitener2.inverse_transform_data(input_data2)

        input_data1 = self.pca1.inverse_transform_data(input_data1)
        input_data2 = self.pca2.inverse_transform_data(input_data2)

        scores1 = self.data["scores1"]
        scores2 = self.data["scores2"]

        patterns1, pvals1 = pearson_correlation(
            input_data1,
            scores2,
            correction=correction,
            alpha=alpha,
            sample_name=self.sample_name,
            feature_name=self.feature_name[0],
        )
        patterns2, pvals2 = pearson_correlation(
            input_data2,
            scores1,
            correction=correction,
            alpha=alpha,
            sample_name=self.sample_name,
            feature_name=self.feature_name[1],
        )

        patterns1.name = "left_heterogeneous_patterns"
        patterns2.name = "right_heterogeneous_patterns"

        pvals1.name = "pvalues_of_left_heterogeneous_patterns"
        pvals2.name = "pvalues_of_right_heterogeneous_patterns"

        patterns1 = self.preprocessor1.inverse_transform_components(patterns1)
        patterns2 = self.preprocessor2.inverse_transform_components(patterns2)

        pvals1 = self.preprocessor1.inverse_transform_components(pvals1)
        pvals2 = self.preprocessor2.inverse_transform_components(pvals2)

        return (patterns1, patterns2), (pvals1, pvals2)

    def _validate_loaded_data(self, data: xr.DataArray):
        if data.attrs.get("placeholder"):
            warnings.warn(
                f"The input data field '{data.name}' was not saved, which will produce"
                " empty results when calling `homogeneous_patterns()` or "
                "`heterogeneous_patterns()`. To avoid this warning, you can save the"
                " model with `save_data=True`, or add the data manually by running"
                " it through the model's `preprocessor.transform()` method and then"
                " attaching it with `data.add()`."
            )

    def _compute_cross_matrix(
        self,
        X: DataArray,
        Y: DataArray,
        sample_dim: str,
        feature_dim_x: str,
        feature_dim_y: str,
        method: str = "covariance",
        diagonal: bool = False,
    ) -> DataArray:
        """Compute the cross matrix of two data objects.

        Assume centered data.

        Parameters
        ----------
        X, Y : DataArray
            DataArrays to compute the cross matrix from.
        sample_dim : str
            Name of the sample dimension.
        feature_dim_x, feature_dim_y : str
            Name of the feature dimensions. If the feature dimensions are the same, they are renamed to avoid conflicts.
        method : {"covariance", "correlation"}
            Method to compute the cross matrix.
        diagonal : bool, default=False
            Whether to compute the diagonal of the cross matrix.

        Returns
        -------
        DataArray
            The cross matrix of the two data objects.

        """
        if feature_dim_x == feature_dim_y:
            new_feature_dim_x = feature_dim_x + "_x"
            new_feature_dim_y = feature_dim_y + "_y"
            X = X.rename({feature_dim_x: new_feature_dim_x})
            Y = Y.rename({feature_dim_y: new_feature_dim_y})
            feature_dim_x = new_feature_dim_x
            feature_dim_y = new_feature_dim_y

        # Rename the sample dimension to avoid conflicts for
        # different coordinates with same length
        sample_dim_x = sample_dim + "_x"
        sample_dim_y = sample_dim + "_y"
        X = X.rename({sample_dim: sample_dim_x})
        Y = Y.rename({sample_dim: sample_dim_y})

        if method == "correlation":
            X = self._normalize_data(X, sample_dim_x)
            Y = self._normalize_data(Y, sample_dim_y)

        if diagonal:
            return xr.apply_ufunc(
                self._compute_cross_covariance_diagonal_numpy,
                X,
                Y,
                input_core_dims=[
                    [sample_dim_x, feature_dim_x],
                    [sample_dim_y, feature_dim_y],
                ],
                output_core_dims=[[feature_dim_y]],
                dask="allowed",
            ).rename({feature_dim_y: feature_dim_y[:-2]})
        else:
            return xr.apply_ufunc(
                self._compute_cross_covariance_numpy,
                X,
                Y,
                input_core_dims=[
                    [sample_dim_x, feature_dim_x],
                    [sample_dim_y, feature_dim_y],
                ],
                output_core_dims=[[feature_dim_x, feature_dim_y]],
                dask="allowed",
            )

    def _compute_cross_covariance_diagonal_numpy(self, X, Y):
        # Assume centered data
        return np.diag(self._compute_cross_covariance_numpy(X, Y))

    def _compute_total_squared_covariance(self, C: DataArray) -> DataArray:
        """Compute the total squared covariance.

        Requires the unwhitened covariance matrix which we can obtain by multiplying the whitened covariance matrix with the inverse of the whitening transformation matrix.
        """
        C = self.whitener2.inverse_transform_data(C)
        C = self.whitener1.inverse_transform_data(C.conj().T)
        # Not necessary to conjugate transpose for total squared covariance
        # C = C.conj().T
        return (abs(C) ** 2).sum()

    @staticmethod
    def _compute_total_variance(X: DataArray, dim: str) -> DataArray:
        """Compute the total variance of the centered data."""
        return (X * X.conj()).sum() / (X[dim].size - 1)

    @staticmethod
    def _compute_cross_covariance_numpy(X, Y):
        # Assume centered data
        n_samples_x = X.shape[0]
        n_samples_y = Y.shape[0]
        if n_samples_x != n_samples_y:
            err_msg = f"Both data matrices must have the same number of samples but found {n_samples_x} in the first and {n_samples_y} in the second."
            raise ValueError(err_msg)
        return X.conj().T @ Y / (n_samples_x - 1)

    @staticmethod
    def _normalize_data(X, dim):
        # Assume centered data
        return X / X.std(dim)


class ComplexCPCCA(CPCCA):
    """Complex CPCCA.

    Complex CPCCA extends classical CPCCA [1]_ by examining amplitude-phase
    relationships. It is based on complex-valued fields obtained from a pair of
    variables such as the zonal and meridional components, :math:`U` and
    :math:`V`, of the wind field, leading to complex-valued data matrices:

        .. math::
        X = U_x + iV_x

    and

    .. math::
        Y = U_y + iV_y

    This method solves the following optimization problem:

        :math:`\\max_{q_x, q_y} \\left( q_x^H X^H Y q_y \\right)`

    subject to the constraints:

        :math:`q_x^H (X^HX)^{1-\\alpha_x} q_x = 1, \\quad q_y^H
        (Y^HY)^{1-\\alpha_y} q_y = 1`

    where :math:`H` denotes the conjugate transpose, :math:`X` and :math:`Y` are
    the complex-valued data matrices, and :math:`\\alpha_x` and
    :math:`\\alpha_y` control the degree of whitening applied to the data.

    Parameters
    ----------
    n_modes : int, default=2
        Number of modes to calculate.
    alpha : Sequence[float] | float, default=0.2
        Degree of whitening applied to the data. If float, the same value is
        applied to both data sets.
    padding : Sequence[str] | str | None, default="exp"
        Padding method for the Hilbert transform. Available options are: - None:
        no padding - "exp": exponential decay
    decay_factor : Sequence[float] | float, default=0.2
        Decay factor for the exponential padding.
    standardize : Squence[bool] | bool, default=False
        Whether to standardize the input data. Generally not recommended as
        standardization can be managed by the degree of whitening.
    use_coslat : Sequence[bool] | bool, default=False
        For data on a longitude-latitude grid, whether to correct for varying
        grid cell areas towards the poles by scaling each grid point with the
        square root of the cosine of its latitude.
    use_pca : Sequence[bool] | bool, default=False
        Whether to preprocess each field individually by reducing dimensionality
        through PCA. The cross-covariance matrix is computed in the reduced
        principal component space.
    n_pca_modes : Sequence[int | float | str] | int | float | str, default=0.999
        Number of modes to retain during PCA preprocessing step. If int,
        specifies the exact number of modes; if float, specifies the fraction of
        variance to retain; if "all", all modes are retained.
    pca_init_rank_reduction : Sequence[float] | float, default=0.3
        Relevant when `use_pca=True` and `n_pca_modes` is a float. Specifies the
        initial fraction of rank reduction for faster PCA computation via
        randomized SVD.
    check_nans : Sequence[bool] | bool, default=True
        Whether to check for NaNs in the input data. Set to False for lazy model
        evaluation.
    compute : bool, default=True
        Whether to compute the model elements eagerly. If True, the following
        are computed sequentially: preprocessor scaler, optional NaN checks, SVD
        decomposition, scores, and components.
    random_state : numpy.random.Generator | int | None, default=None
        Seed for the random number generator.
    sample_name : str, default="sample"
        Name for the new sample dimension.
    feature_name : Sequence[str] | str, default="feature"
        Name for the new feature dimension.
    solver : {"auto", "full", "randomized"}
        Solver to use for the SVD computation.
    solver_kwargs : dict, default={}
        Additional keyword arguments passed to the SVD solver function.

    Examples
    --------

    With two DataArrays :math:`u_i` and :math:`v_i` representing the zonal and
    meridional components of the wind field for two different regions :math:`x`
    and :math:`y`, construct

    >>> X = u_x + 1j * v_x
    >>> Y = u_y + 1j * v_y

    and fit the Complex CPCCA model:

    >>> model = ComplexCPCCA(n_modes=5)
    >>> model.fit(X, Y, "time")

    Finally, extract the amplitude and phase patterns:

    >>> amp_x, amp_y = model.components_amplitude()
    >>> phase_x, phase_y = model.components_phase()


    References
    ----------
    .. [1] Swenson, E. Continuum Power CCA: A Unified Approach for Isolating
        Coupled Modes. Journal of Climate 28, 1016–1030 (2015).

    """

    uses_complex = True

    def __init__(
        self,
        n_modes: int = 2,
        alpha: Sequence[float] | float = 0.2,
        standardize: Sequence[bool] | bool = False,
        use_coslat: Sequence[bool] | bool = False,
        check_nans: Sequence[bool] | bool = True,
        use_pca: Sequence[bool] | bool = True,
        n_pca_modes: Sequence[float | int | str] | float | int | str = 0.999,
        pca_init_rank_reduction: Sequence[float] | float = 0.3,
        compute: bool = True,
        sample_name: str = "sample",
        feature_name: Sequence[str] | str = "feature",
        solver: str = "auto",
        random_state: np.random.Generator | int | None = None,
        solver_kwargs: dict = {},
    ):
        CPCCA.__init__(
            self,
            n_modes=n_modes,
            standardize=standardize,
            use_coslat=use_coslat,
            check_nans=check_nans,
            use_pca=use_pca,
            n_pca_modes=n_pca_modes,
            pca_init_rank_reduction=pca_init_rank_reduction,
            alpha=alpha,
            compute=compute,
            sample_name=sample_name,
            feature_name=feature_name,
            solver=solver,
            random_state=random_state,
            solver_kwargs=solver_kwargs,
        )
        self.attrs.update({"model": "Complex CPCCA"})

    def _fit_algorithm(self, X: DataArray, Y: DataArray) -> Self:
        if (not np.iscomplexobj(X)) or (not np.iscomplexobj(Y)):
            warnings.warn(
                "Expected complex-valued data but found real-valued data. For Hilbert model, use corresponding `Hilbert` class."
            )

        return super()._fit_algorithm(X, Y)

    def components_amplitude(self, normalized=True) -> tuple[DataObject, DataObject]:
        """Get the amplitude of the components.

        The amplitudes of the components are defined as

        .. math::
            A_{x, ij} = |p_{x, ij}|
        .. math::
            A_{y, ij} = |p_{y, ij}|

        where :math:`p_{ij}` is the :math:`i`-th entry of the :math:`j`-th
        component and :math:`|\\cdot|` denotes the absolute value.

        Returns
        -------
        tuple[DataObject, DataObject]
            Component amplitudes of :math:`X` and :math:`Y`.

        """
        Px, Py = self._get_components(normalized=normalized)

        Px = self.whitener1.inverse_transform_components(Px)
        Py = self.whitener2.inverse_transform_components(Py)

        Px = self.pca1.inverse_transform_components(Px)
        Py = self.pca2.inverse_transform_components(Py)

        Px = abs(Px)
        Py = abs(Py)

        Px.name = "components_amplitude_X"
        Py.name = "components_amplitude_Y"

        Px = self.preprocessor1.inverse_transform_components(Px)
        Py = self.preprocessor2.inverse_transform_components(Py)

        return Px, Py

    def components_phase(self, normalized=True) -> tuple[DataObject, DataObject]:
        """Get the phase of the components.

        The phases of the components are defined as

        .. math::
            \\phi_{x, ij} = \\arg(p_{x, ij})
        .. math::
            \\phi_{y, ij} = \\arg(p_{y, ij})

        where :math:`p_{ij}` is the :math:`i`-th entry of the :math:`j`-th component and
        :math:`\\arg(\\cdot)` denotes the argument of a complex number.

        Returns
        -------
        tuple[DataObject, DataObject]
            Component phases of :math:`X` and :math:`Y`.

        """
        Px, Py = self._get_components(normalized=normalized)

        Px = self.whitener1.inverse_transform_components(Px)
        Py = self.whitener2.inverse_transform_components(Py)

        Px = self.pca1.inverse_transform_components(Px)
        Py = self.pca2.inverse_transform_components(Py)

        Px = xr.apply_ufunc(np.angle, Px, keep_attrs=True, dask="allowed")
        Py = xr.apply_ufunc(np.angle, Py, keep_attrs=True, dask="allowed")

        Px.name = "components_phase_X"
        Py.name = "components_phase_Y"

        Px = self.preprocessor1.inverse_transform_components(Px)
        Py = self.preprocessor2.inverse_transform_components(Py)

        return Px, Py

    def scores_amplitude(self, normalized=False) -> tuple[DataArray, DataArray]:
        """Get the amplitude of the scores.

        The amplitudes of the scores are defined as

        .. math::
            A_{x, ij} = |r_{y, ij}|
        .. math::
            A_{y, ij} = |r_{x, ij}|

        where :math:`r_{ij}` is the :math:`i`-th entry of the :math:`j`-th score and
        :math:`|\\cdot|` denotes the absolute value.

        Returns
        -------
        tuple[DataArray, DataArray]
            Score amplitudes of :math:`X` and :math:`Y`.

        """
        Rx, Ry = self._get_scores(normalized=normalized)

        Rx = self.whitener1.inverse_transform_scores(Rx)
        Ry = self.whitener2.inverse_transform_scores(Ry)

        Rx = self.pca1.inverse_transform_scores(Rx)
        Ry = self.pca2.inverse_transform_scores(Ry)

        Rx = abs(Rx)
        Ry = abs(Ry)

        Rx.name = "scores_amplitude_X"
        Ry.name = "scores_amplitude_Y"

        Rx = self.preprocessor1.inverse_transform_scores(Rx)
        Ry = self.preprocessor2.inverse_transform_scores(Ry)

        return Rx, Ry

    def scores_phase(self, normalized=False) -> tuple[DataArray, DataArray]:
        """Get the phase of the scores.

        The phases of the scores are defined as

        .. math::
            \\phi_{x, ij} = \\arg(r_{x, ij})
        .. math::
            \\phi_{y, ij} = \\arg(r_{y, ij})

        where :math:`r_{ij}` is the :math:`i`-th entry of the :math:`j`-th score and
        :math:`\\arg(\\cdot)` denotes the argument of a complex number.

        Returns
        -------
        tuple[DataArray, DataArray]
            Score phases of :math:`X` and :math:`Y`.

        """
        Rx, Ry = self._get_scores(normalized=normalized)

        Rx = self.whitener1.inverse_transform_scores(Rx)
        Ry = self.whitener2.inverse_transform_scores(Ry)

        Rx = self.pca1.inverse_transform_scores(Rx)
        Ry = self.pca2.inverse_transform_scores(Ry)

        Rx = xr.apply_ufunc(np.angle, Rx, keep_attrs=True, dask="allowed")
        Ry = xr.apply_ufunc(np.angle, Ry, keep_attrs=True, dask="allowed")

        Rx.name = "scores_phase_X"
        Ry.name = "scores_phase_Y"

        Rx = self.preprocessor1.inverse_transform_scores(Rx)
        Ry = self.preprocessor2.inverse_transform_scores(Ry)

        return Rx, Ry


class HilbertCPCCA(ComplexCPCCA):
    """Hilbert CPCCA.

    Hilbert CPCCA extends classical CPCCA [1]_ by examining
    amplitude-phase relationships. It augments the input data with its Hilbert
    transform, creating a complex-valued field.

    This method solves the following optimization problem:

        :math:`\\max_{q_x, q_y} \\left( q_x^H X^H Y q_y \\right)`

    subject to the constraints:

        :math:`q_x^H (X^HX)^{1-\\alpha_x} q_x = 1, \\quad q_y^H
        (Y^HY)^{1-\\alpha_y} q_y = 1`

    where :math:`H` denotes the conjugate transpose, :math:`X` and :math:`Y` are
    the augmented data matrices, and :math:`\\alpha_x` and :math:`\\alpha_y`
    control the degree of whitening applied to the data.

    Parameters
    ----------
    n_modes : int, default=2
        Number of modes to calculate.
    alpha : Sequence[float] | float, default=0.2
        Degree of whitening applied to the data. If float, the same value is
        applied to both data sets.
    padding : Sequence[str] | str | None, default="exp"
        Padding method for the Hilbert transform. Available options are: - None:
        no padding - "exp": exponential decay
    decay_factor : Sequence[float] | float, default=0.2
        Decay factor for the exponential padding.
    standardize : Squence[bool] | bool, default=False
        Whether to standardize the input data. Generally not recommended as
        standardization can be managed by the degree of whitening.
    use_coslat : Sequence[bool] | bool, default=False
        For data on a longitude-latitude grid, whether to correct for varying
        grid cell areas towards the poles by scaling each grid point with the
        square root of the cosine of its latitude.
    use_pca : Sequence[bool] | bool, default=False
        Whether to preprocess each field individually by reducing dimensionality
        through PCA. The cross-covariance matrix is computed in the reduced
        principal component space.
    n_pca_modes : Sequence[int | float | str] | int | float | str, default=0.999
        Number of modes to retain during PCA preprocessing step. If int,
        specifies the exact number of modes; if float, specifies the fraction of
        variance to retain; if "all", all modes are retained.
    pca_init_rank_reduction : Sequence[float] | float, default=0.3
        Relevant when `use_pca=True` and `n_pca_modes` is a float. Specifies the
        initial fraction of rank reduction for faster PCA computation via
        randomized SVD.
    check_nans : Sequence[bool] | bool, default=True
        Whether to check for NaNs in the input data. Set to False for lazy model
        evaluation.
    compute : bool, default=True
        Whether to compute the model elements eagerly. If True, the following
        are computed sequentially: preprocessor scaler, optional NaN checks, SVD
        decomposition, scores, and components.
    random_state : numpy.random.Generator | int | None, default=None
        Seed for the random number generator.
    sample_name : str, default="sample"
        Name for the new sample dimension.
    feature_name : Sequence[str] | str, default="feature"
        Name for the new feature dimension.
    solver : {"auto", "full", "randomized"}
        Solver to use for the SVD computation.
    solver_kwargs : dict, default={}
        Additional keyword arguments passed to the SVD solver function.

    Examples
    --------

    Perform Hilbert CPCCA on two real-valued datasets `X` and `Y`, using
    exponential padding:

    >>> model = HilbertCPCCA(n_modes=5, padding="exp")
    >>> model.fit(X, Y)

    References
    ----------
    .. [1] Swenson, E. Continuum Power CCA: A Unified Approach for Isolating
        Coupled Modes. Journal of Climate 28, 1016–1030 (2015).

    """

    def __init__(
        self,
        n_modes: int = 2,
        alpha: Sequence[float] | float = 0.2,
        padding: Sequence[str] | str | None = "exp",
        decay_factor: Sequence[float] | float = 0.2,
        standardize: Sequence[bool] | bool = False,
        use_coslat: Sequence[bool] | bool = False,
        check_nans: Sequence[bool] | bool = True,
        use_pca: Sequence[bool] | bool = True,
        n_pca_modes: Sequence[float | int | str] | float | int | str = 0.999,
        pca_init_rank_reduction: Sequence[float] | float = 0.3,
        compute: bool = True,
        sample_name: str = "sample",
        feature_name: Sequence[str] | str = "feature",
        solver: str = "auto",
        random_state: np.random.Generator | int | None = None,
        solver_kwargs: dict = {},
    ):
        ComplexCPCCA.__init__(
            self,
            n_modes=n_modes,
            standardize=standardize,
            use_coslat=use_coslat,
            check_nans=check_nans,
            use_pca=use_pca,
            n_pca_modes=n_pca_modes,
            pca_init_rank_reduction=pca_init_rank_reduction,
            alpha=alpha,
            compute=compute,
            sample_name=sample_name,
            feature_name=feature_name,
            solver=solver,
            random_state=random_state,
            solver_kwargs=solver_kwargs,
        )
        self.attrs.update({"model": "Hilbert CPCCA"})

        padding = self._process_parameter("padding", padding, "epx")
        decay_factor = self._process_parameter("decay_factor", decay_factor, 0.2)
        self._params["padding"] = padding
        self._params["decay_factor"] = decay_factor

    def _fit_algorithm(self, X: DataArray, Y: DataArray) -> Self:
        CPCCA._fit_algorithm(self, X, Y)
        return self

    def transform(
        self, X: DataObject | None = None, Y: DataObject | None = None, normalized=False
    ) -> Sequence[DataArray]:
        """Transform the input data into the component space."""
        raise NotImplementedError("Hilbert models do not support the transform method.")

    def _augment_data(self, X: DataArray, Y: DataArray) -> tuple[DataArray, DataArray]:
        """Augment the data with the Hilbert transform."""
        params = self.get_params()
        padding = params["padding"]
        decay_factor = params["decay_factor"]
        X = hilbert_transform(
            X,
            dims=(self.sample_name, self.feature_name[0]),
            padding=padding[0],
            decay_factor=decay_factor[0],
        )
        Y = hilbert_transform(
            Y,
            dims=(self.sample_name, self.feature_name[1]),
            padding=padding[1],
            decay_factor=decay_factor[1],
        )
        return X, Y



import numpy as np
import xarray as xr
from typing_extensions import Self

from xeofs.utils.data_types import DataArray

from ..utils.constants import (
    VALID_CARTESIAN_X_NAMES,
    VALID_CARTESIAN_Y_NAMES,
    VALID_LATITUDE_NAMES,
    VALID_LONGITUDE_NAMES,
    VALID_KERNELS,
    VALID_METRICS,
)
from ..utils.sanity_checks import assert_not_complex
from .base_model_single_set import BaseModelSingleSet


class GWPCA(BaseModelSingleSet):
    """Geographically weighted PCA.

    Geographically weighted PCA (GWPCA) [1]_ uses a geographically weighted approach to perform PCA for
    each observation in the dataset based on its local neighbors.

    The neighbors for each observation are determined based on the provided
    bandwidth and metric. Each neighbor is weighted based on its distance from
    the observation using the provided kernel function.

    Parameters
    ----------
    n_modes: int
        Number of modes to calculate.
    bandwidth: float
        Bandwidth of the kernel function. Must be > 0.
    metric: str, default="haversine"
        Distance metric to use. Great circle distance (`haversine`) is always expressed in kilometers.
        All other distance metrics are reported in the unit of the input data.
        See scipy.spatial.distance.cdist for a list of available metrics.
    kernel: str, default="bisquare"
        Kernel function to use. Must be one of ['bisquare', 'gaussian', 'exponential'].
    center: bool, default=True
        If True, the data is centered by subtracting the mean (feature-wise).
    standardize: bool, default=False
        If True, the data is divided by the standard deviation (feature-wise).
    use_coslat: bool, default=False
        If True, the data is weighted by the square root of cosine of latitudes.
    sample_name: str, default="sample"
        Name of the sample dimension.
    feature_name: str, default="feature"
        Name of the feature dimension.

    Attributes
    ----------
    bandwidth: float
        Bandwidth of the kernel function.
    metric: str
        Distance metric to use.
    kernel: str
        Kernel function to use.

    Methods:
    --------
    fit(X) : Fit the model with input data.

    explained_variance() : Return the explained variance of the local components.

    explained_variance_ratio() : Return the explained variance ratio of the local components.

    largest_locally_weighted_components() : Return the largest locally weighted components.


    Notes
    -----
    GWPCA is computationally expensive since it performs PCA for each sample. This implementation leverages
    `numba` to speed up the computation on CPUs. However, for moderate to large datasets, this won't be sufficient.
    Currently, GPU support is not implemented. If you're dataset is too large to be processed on a CPU, consider
    using the R package `GWmodel` [2]_, which provides a GPU implementation of GWPCA.

    References
    ----------
    .. [1] Harris, P., Brunsdon, C. & Charlton, M. Geographically weighted principal components analysis. International Journal of Geographical Information Science 25, 1717–1736 (2011).
    .. [2] https://cran.r-project.org/web/packages/GWmodel/index.html


    """

    extra_modules = ["numba"]

    def __init__(
        self,
        n_modes: int,
        bandwidth: float,
        metric: str = "haversine",
        kernel: str = "bisquare",
        center: bool = True,
        standardize: bool = False,
        use_coslat: bool = False,
        check_nans: bool = True,
        sample_name: str = "sample",
        feature_name: str = "feature",
    ):
        super().__init__(
            n_modes,
            center=center,
            standardize=standardize,
            use_coslat=use_coslat,
            check_nans=check_nans,
            sample_name=sample_name,
            feature_name=feature_name,
        )

        self.attrs.update({"model": "GWPCA"})

        if kernel not in VALID_KERNELS:
            raise ValueError(
                f"Invalid kernel: {kernel}. Must be one of {VALID_KERNELS}."
            )

        if metric not in VALID_METRICS:
            raise ValueError(
                f"Invalid metric: {metric}. Must be one of {VALID_METRICS}."
            )

        if bandwidth <= 0:
            raise ValueError(f"Invalid bandwidth: {bandwidth}. Must be > 0.")

        self.bandwidth = bandwidth
        self.metric = metric
        self.kernel = kernel

    def _fit_algorithm(self, X: DataArray) -> Self:
        # Hide numba imports here to greatly speed up module import time
        from ..utils.optional.numba_utils import _local_pcas

        # Check input type
        assert_not_complex(X)

        # Convert Dask arrays
        if not isinstance(X.data, np.ndarray):
            print(
                "Warning: GWPCA currently does not support Dask arrays. Data is being loaded into memory."
            )
            X = X.compute()
        # 1. Get sample coordinates
        valid_x_names = VALID_CARTESIAN_X_NAMES + VALID_LONGITUDE_NAMES
        valid_y_names = VALID_CARTESIAN_Y_NAMES + VALID_LATITUDE_NAMES
        n_sample_dims = len(self.sample_dims)
        if n_sample_dims == 1:
            indexes = self.preprocessor.preconverter.transformers[0].coords_from_fit
            sample_dims = self.preprocessor.renamer.transformers[0].sample_dims_after
            xy = None
            for dim in sample_dims:
                keys = [k for k in indexes[dim].coords.keys()]
                x_found = any([k.lower() in valid_x_names for k in keys])
                y_found = any([k.lower() in valid_y_names for k in keys])
                if x_found and y_found:
                    xy = np.asarray([*indexes[dim].values])
                    break
            if xy is None:
                raise ValueError("Cannot find sample coordinates.")
        elif n_sample_dims == 2:
            indexes = self.preprocessor.postconverter.transformers[0].coords_from_fit
            xy = np.asarray([*indexes[self.sample_name].values])

        else:
            raise ValueError(
                "GWPCA requires number of sample dimensions to be <= 2, but got {n_sample_dims}."
            )

        # 2. Remove NaN samples from sample indexes
        is_no_nan_sample = self.preprocessor.sanitizer.transformers[0].is_valid_sample
        xy = xr.DataArray(
            xy,
            dims=[self.sample_name, "xy"],
            coords={
                self.sample_name: is_no_nan_sample[self.sample_name],
                "xy": ["x", "y"],
            },
            name="index",
        )

        xy = xy[is_no_nan_sample]

        # Iterate over all samples
        kwargs = {
            "n_modes": self.n_modes,
            "metric": self.metric,
            "kernel": self.kernel,
            "bandwidth": self.bandwidth,
        }
        components, exp_var, tot_var = xr.apply_ufunc(
            _local_pcas,
            X,
            xy,
            input_core_dims=[
                [self.sample_name, self.feature_name],
                [self.sample_name, "xy"],
            ],
            output_core_dims=[
                [self.sample_name, self.feature_name, "mode"],
                [self.sample_name, "mode"],
                [self.sample_name],
            ],
            kwargs=kwargs,
            dask="forbidden",
        )
        components = components.assign_coords(
            {
                self.sample_name: X[self.sample_name],
                self.feature_name: X[self.feature_name],
                "mode": np.arange(1, self.n_modes + 1),
            }
        )
        exp_var = exp_var.assign_coords(
            {
                self.sample_name: X[self.sample_name],
                "mode": np.arange(1, self.n_modes + 1),
            }
        )
        tot_var = tot_var.assign_coords({self.sample_name: X[self.sample_name]})

        exp_var_ratio = exp_var / tot_var

        # self.data.add(X, "input_data")
        self.data.add(components, "components")
        self.data.add(exp_var, "explained_variance")
        self.data.add(exp_var_ratio, "explained_variance_ratio")

        self.data.set_attrs(self.attrs)

        return self

    def explained_variance(self):
        expvar = self.data["explained_variance"]
        return self.preprocessor.inverse_transform_scores(expvar)

    def explained_variance_ratio(self):
        expvar = self.data["explained_variance_ratio"]
        return self.preprocessor.inverse_transform_scores(expvar)

    def largest_locally_weighted_components(self):
        comps = self.data["components"]
        idx_max = abs(comps).argmax(self.feature_name)
        input_features = self.preprocessor.stacker.transformers[0].coords_out["feature"]
        llwc = input_features[idx_max].drop_vars(self.feature_name)
        llwc.name = "largest_locally_weighted_components"
        return self.preprocessor.inverse_transform_scores(llwc)

    def scores(self):
        raise NotImplementedError("GWPCA does not support scores() yet.")

    def _transform_algorithm(self, data: DataArray) -> DataArray:
        raise NotImplementedError("GWPCA does not support transform() yet.")

    def _inverse_transform_algorithm(self, scores: DataArray) -> DataArray:
        raise NotImplementedError("GWPCA does not support inverse_transform() yet.")



"""
Complex EOF analysis
============================================

In this tutorial, we'll walk through how to perform a Complex EOF analysis on
the zonal and meridional wind components.

Let's start by importing the necessary packages and loading the data:
"""

# %%
import matplotlib.pyplot as plt
import xarray as xr

import xeofs as xe

xr.set_options(display_expand_attrs=False)

# %%
# For this example, we'll use the ERA-Interim tutorial dataset ``eraint_uvz``:

uvz = xr.tutorial.open_dataset("eraint_uvz")
uvz

# %%
# This dataset contains the zonal, meridional, and vertical wind components at
# three different atmospheric levels. Note that the data only covers two months,
# so we have just two time steps (samples). While this isn't enough for a robust
# EOF analysis, we'll proceed for demonstration purposes. Now, let's combine the
# zonal (``u``) and meridional (``v``) wind components into a complex-valued
# dataset:

Z = uvz["u"] + 1j * uvz["v"]

# %%
# Next, we'll initialize and fit the ``ComplexEOF`` model to our data. The
# ``xeofs`` package makes this easy by allowing us to specify the sample
# dimension (``month``), automatically performing the Complex EOF analysis
# across all three atmospheric levels. As a standard practice, we'll also weigh
# each grid cell by the square root of the cosine of the latitude
# (``use_coslat=True``).

model = xe.single.ComplexEOF(n_modes=1, use_coslat=True, random_state=7)
model.fit(Z, dim="month")

# %%
# Instead of just extracting the complex-valued components, we can also get the
# amplitude and phase of these components. Let's start by looking at the
# amplitude of the first mode:


spatial_ampltiudes = model.components_amplitude()
spatial_phases = model.components_phase()

spatial_ampltiudes.sel(mode=1).plot(col="level")
plt.show()

# %%
# It looks like the first mode picks up a pattern resembling the location of the
# subtropical jet stream around ±30º latitude, particularly strong in the upper
# troposphere at 200 hPa and weaker toward the surface. We can also plot the
# phase of the first mode. To keep the plot clear, we'll only show the phase
# where the amplitude is above a certain threshold (e.g., 0.004):

relevant_phases = spatial_phases.where(spatial_ampltiudes > 0.004)
relevant_phases.sel(mode=1).plot(col="level", cmap="twilight")
plt.show()

# %%



import pytest
import xarray as xr
import numpy as np

from xeofs.preprocessing.scaler import Scaler


@pytest.mark.parametrize(
    "with_std, with_coslat",
    [
        (True, True),
        (True, False),
        (False, True),
        (False, False),
    ],
)
def test_init_params(with_std, with_coslat):
    s = Scaler(with_std=with_std, with_coslat=with_coslat)
    assert s.get_params()["with_std"] == with_std
    assert s.get_params()["with_coslat"] == with_coslat


@pytest.mark.parametrize(
    "with_std, with_coslat",
    [
        (True, True),
        (True, False),
        (False, True),
        (False, False),
    ],
)
def test_fit_params(with_std, with_coslat, mock_dataset):
    s = Scaler(with_std=with_std, with_coslat=with_coslat)
    sample_dims = ["time"]
    feature_dims = ["lat", "lon"]
    size_lats = mock_dataset.lat.size
    weights = xr.DataArray(
        np.random.rand(size_lats), dims=["lat"], name="weights"
    ).to_dataset()
    s.fit(mock_dataset, sample_dims, feature_dims, weights)
    assert hasattr(s, "mean_"), "Scaler has no mean attribute."
    if with_std:
        assert hasattr(s, "std_"), "Scaler has no std attribute."
    if with_coslat:
        assert hasattr(s, "coslat_weights_"), "Scaler has no coslat_weights attribute."
    assert s.mean_ is not None, "Scaler mean is None."
    if with_std:
        assert s.std_ is not None, "Scaler std is None."
    if with_coslat:
        assert s.coslat_weights_ is not None, "Scaler coslat_weights is None."


@pytest.mark.parametrize(
    "with_std, with_coslat, with_weights",
    [
        (True, True, True),
        (True, False, True),
        (False, True, True),
        (False, False, True),
        (True, True, False),
        (True, False, False),
        (False, True, False),
        (False, False, False),
    ],
)
def test_transform_params(with_std, with_coslat, with_weights, mock_dataset):
    s = Scaler(with_std=with_std, with_coslat=with_coslat)
    sample_dims = ["time"]
    feature_dims = ["lat", "lon"]
    size_lats = mock_dataset.lat.size
    if with_weights:
        weights1 = xr.DataArray(np.random.rand(size_lats), dims=["lat"], name="t2m")
        weights2 = xr.DataArray(np.random.rand(size_lats), dims=["lat"], name="prcp")
        weights = xr.merge([weights1, weights2])
    else:
        weights = None
    s.fit(mock_dataset, sample_dims, feature_dims, weights)
    transformed = s.transform(mock_dataset)
    assert transformed is not None, "Transformed data is None."

    transformed_mean = transformed.mean(sample_dims, skipna=False)
    assert np.allclose(
        transformed_mean.to_array(), 0
    ), "Mean of the transformed data is not zero."

    if with_std:
        transformed_std = transformed.std(sample_dims, skipna=False)
        if with_coslat or with_weights:
            assert (
                transformed_std <= 1
            ).all(), "Standard deviation of the transformed data is larger one."
        else:
            assert np.allclose(
                transformed_std.to_array(), 1
            ), "Standard deviation of the transformed data is not one."

    if with_coslat:
        assert s.coslat_weights_ is not None, "Scaler coslat_weights is None."
        assert not np.array_equal(
            transformed, mock_dataset
        ), "Data has not been transformed."

    transformed2 = s.fit_transform(mock_dataset, sample_dims, feature_dims, weights)
    xr.testing.assert_allclose(transformed, transformed2)


@pytest.mark.parametrize(
    "with_std, with_coslat",
    [
        (True, True),
        (True, False),
        (False, True),
        (False, False),
    ],
)
def test_inverse_transform_params(with_std, with_coslat, mock_dataset):
    s = Scaler(with_std=with_std, with_coslat=with_coslat)
    sample_dims = ["time"]
    feature_dims = ["lat", "lon"]
    size_lats = mock_dataset.lat.size
    weights1 = xr.DataArray(np.random.rand(size_lats), dims=["lat"], name="t2m")
    weights2 = xr.DataArray(np.random.rand(size_lats), dims=["lat"], name="prcp")
    weights = xr.merge([weights1, weights2])
    s.fit(mock_dataset, sample_dims, feature_dims, weights)
    transformed = s.transform(mock_dataset)
    inverted = s.inverse_transform_data(transformed)
    xr.testing.assert_allclose(inverted, mock_dataset)


@pytest.mark.parametrize(
    "dim_sample, dim_feature",
    [
        (("time",), ("lat", "lon")),
        (("time",), ("lon", "lat")),
        (("lat", "lon"), ("time",)),
        (("lon", "lat"), ("time",)),
    ],
)
def test_fit_dims(dim_sample, dim_feature, mock_dataset):
    s = Scaler(with_std=True)
    s.fit(mock_dataset, dim_sample, dim_feature)
    assert hasattr(s, "mean_"), "Scaler has no mean attribute."
    assert s.mean_ is not None, "Scaler mean is None."
    assert hasattr(s, "std_"), "Scaler has no std attribute."
    assert s.std_ is not None, "Scaler std is None."
    # check that all dimensions are present except the sample dimensions
    assert set(s.mean_.dims) == set(mock_dataset.dims) - set(
        dim_sample
    ), "Mean has wrong dimensions."
    assert set(s.std_.dims) == set(mock_dataset.dims) - set(
        dim_sample
    ), "Standard deviation has wrong dimensions."


@pytest.mark.parametrize(
    "dim_sample, dim_feature",
    [
        (("time",), ("lat", "lon")),
        (("time",), ("lon", "lat")),
        (("lat", "lon"), ("time",)),
        (("lon", "lat"), ("time",)),
    ],
)
def test_fit_transform_dims(dim_sample, dim_feature, mock_dataset):
    s = Scaler()
    transformed = s.fit_transform(mock_dataset, dim_sample, dim_feature)
    # check that all dimensions are present
    assert set(transformed.dims) == set(
        mock_dataset.dims
    ), "Transformed data has wrong dimensions."
    # check that the coordinates are the same
    for dim in mock_dataset.dims:
        xr.testing.assert_allclose(transformed[dim], mock_dataset[dim])


# Test input types
def test_fit_input_type(mock_dataset, mock_data_array, mock_data_array_list):
    s = Scaler()
    # Cannot fit list of DataArrays
    with pytest.raises(TypeError):
        s.fit(mock_data_array_list, ["time"], ["lon", "lat"])

    s.fit(mock_dataset, ["time"], ["lon", "lat"])

    # Cannot transform list of DataArrays
    with pytest.raises(TypeError):
        s.transform(mock_data_array_list)


# def test_fit_weights_input_type(mock_dataset):
#     s = Scaler()
#     # Fitting with weights requires that the weights have the same variables as the dataset
#     # used for fitting; otherwise raise an error
#     size_lats = mock_dataset.lat.size
#     weights1 = xr.DataArray(np.random.rand(size_lats), dims=['lat'], name='t2m')  # correct name
#     weights2 = xr.DataArray(np.random.rand(size_lats), dims=['lat'], name='prcp') # correct name
#     weights3 = xr.DataArray(np.random.rand(size_lats), dims=['lat'], name='sic') # wrong name
#     weights_correct = xr.merge([weights1, weights2])
#     weights_wrong = xr.merge([weights1, weights3])

#     s.fit(mock_dataset, ['time'], ['lon', 'lat'], weights_correct)

#     with pytest.raises(ValueError):
#         s.fit(mock_dataset, ['time'], ['lon', 'lat'], weights_wrong)



import pytest
import numpy as np
import pandas as pd
from xeofs.utils.data_types import (
    DataArray,
    DataSet,
    DataList,
    Dims,
    DimsList,
    DimsTuple,
    DimsListTuple,
)
from xeofs.utils.xarray_utils import data_is_dask  # noqa: F401


def is_xdata(data):
    return isinstance(data, (DataArray, DataSet))


def get_dims_from_data(data: DataArray | DataSet) -> DimsTuple:
    # If data is DataArray/Dataset
    if is_xdata(data):
        data_dims: Dims = tuple(data.dims)
        sample_dims: Dims = tuple([dim for dim in data.dims if "sample" in str(dim)])
        feature_dims: Dims = tuple([dim for dim in data.dims if "feature" in str(dim)])
        return data_dims, sample_dims, feature_dims
    else:
        raise ValueError("unrecognized input type")


def get_dims_from_data_list(data_list: DataList) -> DimsListTuple:
    # If data is list
    if isinstance(data_list, list):
        data_dims: DimsList = [data.dims for data in data_list]
        sample_dims: DimsList = []
        feature_dims: DimsList = []
        for data in data_list:
            sdims = tuple([dim for dim in data.dims if "sample" in str(dim)])
            fdims = tuple([dim for dim in data.dims if "feature" in str(dim)])
            sample_dims.append(sdims)
            feature_dims.append(fdims)
        return data_dims, sample_dims, feature_dims

    else:
        raise ValueError("unrecognized input type")


def data_has_multiindex(data: DataArray | DataSet | DataList) -> bool:
    """Check if the given data object has any MultiIndex."""
    if isinstance(data, DataArray) or isinstance(data, DataSet):
        return any(isinstance(index, pd.MultiIndex) for index in data.indexes.values())
    elif isinstance(data, list):
        return all(data_has_multiindex(da) for da in data)
    else:
        raise ValueError("unrecognized input type")


def assert_expected_dims(data1, data2, policy="all"):
    """
    Check if dimensions of two data objects matches.

    Parameters:
    - data1: Reference data object (either a DataArray, DataSet, or list of DataArray)
    - data2: Test data object (same type as data1)
    - policy: Policy to check the dimensions. Can be either "all", "feature" or "sample"

    """

    if is_xdata(data1) and is_xdata(data2):
        all_dims1, sample_dims1, feature_dims1 = get_dims_from_data(data1)
        all_dims2, sample_dims2, feature_dims2 = get_dims_from_data(data2)

        if policy == "all":
            err_msg = "Dimensions do not match: {:} vs {:}".format(all_dims1, all_dims2)
            assert set(all_dims1) == set(all_dims2), err_msg
        elif policy == "feature":
            err_msg = "Dimensions do not match: {:} vs {:}".format(
                feature_dims1, feature_dims2
            )
            assert set(feature_dims1) == set(feature_dims2), err_msg
            assert len(sample_dims2) == 0, "Sample dimensions should be empty"
            assert "mode" in all_dims2, "Mode dimension is missing"

        elif policy == "sample":
            err_msg = "Dimensions do not match: {:} vs {:}".format(
                sample_dims1, sample_dims2
            )
            assert set(sample_dims1) == set(sample_dims2), err_msg
            assert len(feature_dims2) == 0, "Feature dimensions should be empty"
            assert "mode" in all_dims2, "Mode dimension is missing"
        else:
            raise ValueError("Unrecognized policy: {:}".format(policy))

    elif isinstance(data1, list) and isinstance(data2, list):
        for da1, da2 in zip(data1, data2):
            assert_expected_dims(da1, da2, policy=policy)

    # If neither of the above conditions are met, raise an error
    else:
        raise ValueError(
            "Cannot check coordinates. Unrecognized data type. data1: {:}, data2: {:}".format(
                type(data1), type(data2)
            )
        )


def assert_expected_coords(data1, data2, policy="all") -> None:
    """
    Check if coordinates of the data objects matches.

    Parameters:
    - data1: Reference data object (either a DataArray, DataSet, or list of DataArray)
    - data2: Test data object (same type as data1)
    - policy: Policy to check the dimensions. Can be either "all", "feature" or "sample"

    """

    # Data objects is either DataArray or DataSet
    if is_xdata(data1) and is_xdata(data2):
        all_dims1, sample_dims1, feature_dims1 = get_dims_from_data(data1)
        all_dims2, sample_dims2, feature_dims2 = get_dims_from_data(data2)
        if policy == "all":
            assert all(
                np.all(data1.coords[dim].values == data2.coords[dim].values)
                for dim in all_dims1
            )
        elif policy == "feature":
            assert all(
                np.all(data1.coords[dim].values == data2.coords[dim].values)
                for dim in feature_dims1
            )
        elif policy == "sample":
            assert all(
                np.all(data1.coords[dim].values == data2.coords[dim].values)
                for dim in sample_dims1
            )
        else:
            raise ValueError("Unrecognized policy: {:}".format(policy))

    # Data object is list
    elif isinstance(data1, list) and isinstance(data2, list):
        for da1, da2 in zip(data1, data2):
            assert_expected_coords(da1, da2, policy=policy)

    # If neither of the above conditions are met, raise an error
    else:
        raise ValueError(
            "Cannot check coordinates. Unrecognized data type. data1: {:}, data2: {:}".format(
                type(data1), type(data2)
            )
        )


def skip_if_missing_engine(engine: str):
    """
    Skip save/load tests if missing the i/o backend.
    """
    # xarray uses engine="netcdf4" but the package itself is "netCDF4".
    mapping = {"h5netcdf": "h5netcdf", "netcdf4": "netCDF4", "zarr": "zarr"}
    module = mapping.get(engine)
    if module is None:
        raise ValueError(f"Unrecognized engine: {engine}")
    pytest.importorskip(module)
